from __future__ import annotations

import socket
from datetime import datetime
from pathlib import Path

from one_run.models import Manifest, RunMetadata
from one_run.utils.shell import run_capture


def _git_commit(project_root: Path) -> str | None:
    proc = run_capture(["git", "rev-parse", "HEAD"], cwd=project_root)
    if proc.returncode != 0:
        return None
    commit = proc.stdout.strip()
    return commit or None


def _git_dirty(project_root: Path) -> bool | None:
    proc = run_capture(["git", "status", "--porcelain"], cwd=project_root)
    if proc.returncode != 0:
        return None
    return bool(proc.stdout.strip())


def build_metadata(
    run_id: str,
    manifest: Manifest,
    project_root: Path,
    cwd: Path,
    *,
    slurm_job_id: str | None = None,
) -> RunMetadata:
    return RunMetadata(
        run_id=run_id,
        timestamp=datetime.now(),
        hostname=socket.gethostname(),
        cwd=str(cwd),
        project_root=str(project_root),
        git_commit=_git_commit(project_root),
        git_dirty=_git_dirty(project_root),
        experiment_name=manifest.experiment.name,
        tags=manifest.experiment.tags,
        experiment_color=manifest.experiment.color,
        backend_type=manifest.backend.kind,
        docker_image=manifest.environment.image if manifest.environment.kind == "docker" else None,
        docker_image_digest=None,
        slurm_job_id=slurm_job_id,
        slurm_attempt=1 if manifest.backend.kind == "slurm" else None,
    )



from one_run.listeners.manifest import build_routes_from_manifest, manifest_interval_sec
from one_run.listeners.routes import ListenerRoute, load_handler

__all__ = [
    "ListenerRoute",
    "build_routes_from_manifest",
    "manifest_interval_sec",
    "load_handler",
]



from __future__ import annotations

import os
import sys

_ANSI_BY_COLOR = {
    "black": "30",
    "red": "31",
    "green": "32",
    "yellow": "33",
    "blue": "34",
    "magenta": "35",
    "cyan": "36",
    "white": "37",
}


def should_use_color() -> bool:
    if os.getenv("ONE_RUN_NO_COLOR") == "1":
        return False
    if os.getenv("ONE_RUN_FORCE_COLOR") == "1":
        return True
    return sys.stdout.isatty()


def colorize(value: str, color: str | None, *, enabled: bool) -> str:
    if not enabled or not color:
        return value
    code = _ANSI_BY_COLOR.get(color)
    if not code:
        return value
    return f"\x1b[{code}m{value}\x1b[0m"


def render_table(columns: list[str], rows: list[dict[str, str]]) -> list[str]:
    widths = {col: len(col) for col in columns}
    for row in rows:
        for col in columns:
            widths[col] = max(widths[col], len(row[col]))

    def _line(row: dict[str, str]) -> str:
        return "  ".join(row[col].ljust(widths[col]) for col in columns)

    header = "  ".join(col.ljust(widths[col]) for col in columns)
    return [header] + [_line(row) for row in rows]



from __future__ import annotations

import argparse
import os

from one_run import __version__
from one_run.constants import AUTOMATION_ENV_VAR
from one_run.commands.init_cmd import init_listener, init_manifest
from one_run.commands.ls_cmd import list_runs
from one_run.commands.ps_cmd import list_processes
from one_run.commands.cancel_cmd import cancel_run
from one_run.commands.cron_cmd import install_cron, remove_cron, show_cron
from one_run.commands.listen_cmd import listen_events
from one_run.commands.rm_cmd import remove_run, remove_runs
from one_run.commands.run_cmd import run_local
from one_run.commands.status_cmd import show_status
from one_run.commands.submit_cmd import submit_slurm
from one_run.commands.tail_cmd import tail_run_log
from one_run.commands.validate_cmd import validate_manifest
from one_run.errors import OneRunError


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="one-run", description="ml run wrapper for reprodusible experiments and documentation")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument(
        "--automation",
        action="store_true",
        help=f"prefer machine-readable output defaults (also enabled by {AUTOMATION_ENV_VAR}=1)",
    )
    sub = parser.add_subparsers(dest="command", required=True)

    init_p = sub.add_parser("init", help="interactive config generators")
    init_sub = init_p.add_subparsers(dest="init_target", required=True)
    init_manifest_p = init_sub.add_parser("manifest", help="generate a manifest yaml interactively")
    init_manifest_p.add_argument("-o", "--output", default=None, help="output path (default: ./manifests/<experiment>.manifest.yaml)")
    init_manifest_p.add_argument("--force", action="store_true", help="overwrite output file if it exists")
    init_listener_p = init_sub.add_parser("listener", help="generate a listener yaml interactively")
    init_listener_p.add_argument("-o", "--output", default=None, help="output path (default: ./listeners/<name>.listener.yaml)")
    init_listener_p.add_argument("--force", action="store_true", help="overwrite output file if it exists")

    run_p = sub.add_parser("run", help="execute a local run")
    run_p.add_argument("config", help="path to manifest yaml")
    run_p.add_argument("-w", "--watch", action="store_true", help="run in background and return immediately")
    run_p.add_argument("--wait-for", action="append", help="dependency run_id to wait for before start (repeatable)")
    run_p.add_argument("--wait-timeout", help="max dependency wait duration, e.g. 30m")
    run_p.add_argument("--wait-interval", type=float, help="dependency poll interval seconds")
    run_p.add_argument("--json", action="store_true", help="print machine-readable result json")

    submit_p = sub.add_parser("submit", help="submit a slurm run")
    submit_p.add_argument("config", help="path to manifest yaml")
    submit_p.add_argument("--wait-for", action="append", help="dependency run_id to wait for before submit (repeatable)")
    submit_p.add_argument("--wait-timeout", help="max dependency wait duration, e.g. 30m")
    submit_p.add_argument("--wait-interval", type=float, help="dependency poll interval seconds")
    submit_p.add_argument("--json", action="store_true", help="print machine-readable result json")

    status_p = sub.add_parser("status", help="show run status")
    status_p.add_argument("run_id", help="run identifier")
    status_p.add_argument("--json", action="store_true", help="print status payload as json")
    status_p.add_argument("-l", "--long", action="store_true", help="print extended status fields")
    status_p.add_argument("--wait", action="store_true", help="wait until run reaches terminal state")
    status_p.add_argument("--timeout", help="max wait duration, e.g. 30m")
    status_p.add_argument("--interval", type=float, default=2.0, help="poll interval in seconds (default: 2)")

    cancel_p = sub.add_parser("cancel", help="cancel a running run")
    cancel_p.add_argument("run_id", help="run identifier")
    cancel_p.add_argument("--force", action="store_true", help="force kill local run / kill signal for slurm")
    cancel_p.add_argument("--json", action="store_true", help="print machine-readable result json")

    rm_p = sub.add_parser("rm", help="remove runs (single run_id or filtered cleanup)")
    rm_p.add_argument("run_id", nargs="?", help="run identifier")
    rm_p.add_argument("--force", action="store_true", help="force kill when stopping active run")
    rm_p.add_argument("--state", choices=["pending", "running", "success", "failed"], help="filter by run state (bulk mode)")
    rm_p.add_argument("--older-than", help="delete runs older than duration, e.g. 7d (bulk mode)")
    rm_p.add_argument("--dry-run", action="store_true", help="show what would be deleted (bulk mode)")
    rm_p.add_argument("--json", action="store_true", help="print machine-readable result json")

    validate_p = sub.add_parser("validate", help="validate manifest only")
    validate_p.add_argument("config", help="path to manifest yaml")
    validate_p.add_argument("--json", action="store_true", help="print machine-readable result json")

    cron_p = sub.add_parser("cron", help="manage crontab entries from manifest schedule")
    cron_sub = cron_p.add_subparsers(dest="cron_target", required=True)
    cron_install_p = cron_sub.add_parser("install", help="install or update crontab entry for manifest")
    cron_install_p.add_argument("config", help="path to manifest yaml")
    cron_remove_p = cron_sub.add_parser("remove", help="remove crontab entry for manifest")
    cron_remove_p.add_argument("config", help="path to manifest yaml")
    cron_show_p = cron_sub.add_parser("show", help="show crontab entry for manifest")
    cron_show_p.add_argument("config", help="path to manifest yaml")

    tail_p = sub.add_parser("tail", help="show run logs")
    tail_p.add_argument("run_id", help="run identifier")
    tail_p.add_argument("-n", "--lines", type=int, default=30, help="number of lines from the end")
    tail_p.add_argument("-f", "--follow", action="store_true", help="follow log output")
    tail_p.add_argument("--stderr", action="store_true", help="tail stderr instead of stdout")
    tail_p.add_argument("--since", help="show logs only for runs updated within duration (e.g. 10m, 2h)")

    listen_p = sub.add_parser("listen", help="listen to lifecycle events (all runs or selected run_ids)", add_help=False)
    listen_p.add_argument("--help", action="help", help="show this help message and exit")
    listen_p.add_argument("run_ids", nargs="*", help="optional run identifiers")
    listen_p.add_argument("-m", "--manifest", help="manifest with listeners config")
    listen_p.add_argument("--output", choices=["table", "jsonl"], default="table", help="output format (default: table)")
    listen_p.add_argument("--jsonl", action="store_true", help="alias for --output jsonl")
    listen_p.add_argument("--event", action="append", help="filter by event name (repeatable)")
    listen_p.add_argument("--state", action="append", help="filter by event state (repeatable)")
    listen_p.add_argument("--tag", action="append", help="filter by run tag (repeatable)")
    listen_p.add_argument("--handler", help="python callback in module:function format")
    listen_mode_group = listen_p.add_mutually_exclusive_group()
    listen_mode_group.add_argument("--once", action="store_true", help="read currently available events once and exit")
    listen_mode_group.add_argument("-w", "--watch", action="store_true", help="watch event stream continuously")
    listen_p.add_argument("-h", "--history", action="store_true", help="include historical events that happened before listener start")
    listen_p.add_argument("--since", help="history window, e.g. 10m or 2h (requires --history)")
    listen_p.add_argument("--interval", type=float, default=None, help="poll interval in seconds (default: manifest interval_sec or 1.0)")

    ls_p = sub.add_parser("ls", help="list runs")
    ls_p.add_argument("--long", "-l", action="store_true", help="show detailed columns")
    ls_p.add_argument(
        "--state",
        choices=["pending", "running", "success", "failed"],
        help="filter by run state",
    )
    ls_p.add_argument("--tag", action="append", help="filter by tag (can be used multiple times)")
    ls_p.add_argument("--group-by-tag", action="store_true", help="group output by tag")
    ls_p.add_argument("--json", action="store_true", help="print machine-readable json")

    ps_p = sub.add_parser("ps", help="list active runs")
    ps_p.add_argument(
        "--state",
        choices=["pending", "running"],
        help="filter by active state",
    )
    ps_p.add_argument("--tag", action="append", help="filter by tag (can be used multiple times)")
    ps_p.add_argument("--json", action="store_true", help="print machine-readable json")
    ps_p.add_argument("-w", "--watch", action="store_true", help="refresh process table continuously")
    ps_p.add_argument("--interval", type=float, default=2.0, help="watch refresh interval in seconds (default: 2)")

    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    try:
        import argcomplete  # type: ignore

        argcomplete.autocomplete(parser)
    except Exception:
        pass
    args = parser.parse_args(argv)
    automation_mode = args.automation or os.getenv(AUTOMATION_ENV_VAR) == "1"

    try:
        if args.command == "init":
            if args.init_target == "manifest":
                return init_manifest(output=args.output, force=args.force)
            if args.init_target == "listener":
                return init_listener(output=args.output, force=args.force)
        if args.command == "run":
            return run_local(
                args.config,
                watch=args.watch,
                json_output=(args.json or automation_mode),
                wait_for=args.wait_for,
                wait_timeout=args.wait_timeout,
                wait_interval=args.wait_interval,
            )
        if args.command == "submit":
            return submit_slurm(
                args.config,
                json_output=(args.json or automation_mode),
                wait_for=args.wait_for,
                wait_timeout=args.wait_timeout,
                wait_interval=args.wait_interval,
            )
        if args.command == "status":
            return show_status(
                args.run_id,
                json_output=(args.json or automation_mode),
                long=args.long,
                wait=args.wait,
                timeout=args.timeout,
                interval_sec=args.interval,
            )
        if args.command == "cancel":
            return cancel_run(args.run_id, force=args.force, json_output=(args.json or automation_mode))
        if args.command == "rm":
            if args.run_id:
                if args.state or args.older_than or args.dry_run:
                    raise OneRunError("rm <run_id> cannot be combined with --state/--older-than/--dry-run")
                return remove_run(args.run_id, force=args.force, json_output=(args.json or automation_mode))
            return remove_runs(
                state=args.state,
                older_than=args.older_than,
                dry_run=args.dry_run,
                json_output=(args.json or automation_mode),
            )
        if args.command == "validate":
            return validate_manifest(args.config, json_output=(args.json or automation_mode))
        if args.command == "cron":
            if args.cron_target == "install":
                return install_cron(args.config)
            if args.cron_target == "remove":
                return remove_cron(args.config)
            if args.cron_target == "show":
                return show_cron(args.config)
        if args.command == "tail":
            return tail_run_log(args.run_id, stderr=args.stderr, follow=args.follow, lines=args.lines, since=args.since)
        if args.command == "listen":
            watch_mode = args.watch or not args.once
            return listen_events(
                run_ids=args.run_ids,
                watch=watch_mode,
                jsonl=args.jsonl,
                output=args.output,
                interval_sec=args.interval,
                events=args.event,
                states=args.state,
                tags=args.tag,
                handler=args.handler,
                manifest_path=args.manifest,
                once=args.once,
                history=args.history,
                since=args.since,
            )
        if args.command == "ls":
            return list_runs(
                state_filter=args.state,
                long=args.long,
                tags=args.tag,
                group_by_tag=args.group_by_tag,
                json_output=(args.json or automation_mode),
            )
        if args.command == "ps":
            return list_processes(
                state_filter=args.state,
                tags=args.tag,
                json_output=(args.json or automation_mode),
                watch=args.watch,
                interval_sec=args.interval,
            )
    except OneRunError as exc:
        parser.exit(status=1, message=f"error: {exc}\n")

    parser.exit(status=2, message="unknown command\n")
    return 2



from __future__ import annotations

import hashlib
import shlex
import subprocess
import sys
from pathlib import Path

from one_run.errors import OneRunError
from one_run.manifest import load_manifest
from one_run.utils.shell import command_exists


def _cron_marker(manifest_path: Path) -> str:
    digest = hashlib.sha1(str(manifest_path).encode("utf-8")).hexdigest()[:12]
    return f"one-run:{digest}"


def _validate_cron(expr: str) -> None:
    value = expr.strip()
    if not value:
        raise OneRunError("schedule.cron cannot be empty")
    if value.startswith("@"):
        return
    parts = value.split()
    if len(parts) != 5:
        raise OneRunError("schedule.cron must have 5 fields or start with @ (example: @daily)")


def _read_crontab() -> str:
    proc = subprocess.run(["crontab", "-l"], check=False, capture_output=True, text=True)
    if proc.returncode == 0:
        return proc.stdout
    stderr = (proc.stderr or "").lower()
    if "no crontab" in stderr:
        return ""
    raise OneRunError(f"failed to read crontab: {(proc.stderr or proc.stdout).strip()}")


def _write_crontab(text: str) -> None:
    proc = subprocess.run(["crontab", "-"], input=text, check=False, capture_output=True, text=True)
    if proc.returncode != 0:
        raise OneRunError(f"failed to write crontab: {(proc.stderr or proc.stdout).strip()}")


def _remove_marker_lines(text: str, marker: str) -> tuple[str, bool]:
    lines = text.splitlines()
    kept = [line for line in lines if marker not in line]
    removed = len(kept) != len(lines)
    out = "\n".join(kept).strip()
    if out:
        out += "\n"
    return out, removed


def _build_cron_line(*, cron_expr: str, project_root: Path, manifest_path: Path, marker: str, backend_kind: str) -> str:
    action = "submit" if backend_kind == "slurm" else "run"
    python_bin = shlex.quote(sys.executable)
    root = shlex.quote(str(project_root))
    manifest = shlex.quote(str(manifest_path))
    log_path = shlex.quote(str(project_root / "runs" / "cron.log"))
    return f"{cron_expr} cd {root} && {python_bin} -m one_run {action} {manifest} >> {log_path} 2>&1 # {marker}"


def install_cron(config_path: str) -> int:
    if not command_exists("crontab"):
        raise OneRunError("crontab command not found on this system")

    manifest_path = Path(config_path).resolve()
    manifest = load_manifest(manifest_path)
    schedule = manifest.schedule
    if schedule is None:
        raise OneRunError("manifest has no schedule section")
    if not schedule.enabled:
        raise OneRunError("schedule.enabled is false")
    _validate_cron(schedule.cron)

    marker = _cron_marker(manifest_path)
    project_root = Path.cwd().resolve()
    line = _build_cron_line(
        cron_expr=schedule.cron.strip(),
        project_root=project_root,
        manifest_path=manifest_path,
        marker=marker,
        backend_kind=manifest.backend.kind,
    )

    current = _read_crontab()
    cleaned, _ = _remove_marker_lines(current, marker)
    new_text = (cleaned + line + "\n")
    _write_crontab(new_text)

    print(f"installed=true")
    print(f"marker={marker}")
    print(f"manifest={manifest_path}")
    print(f"cron={schedule.cron.strip()}")
    return 0


def remove_cron(config_path: str) -> int:
    if not command_exists("crontab"):
        raise OneRunError("crontab command not found on this system")
    manifest_path = Path(config_path).resolve()
    marker = _cron_marker(manifest_path)
    current = _read_crontab()
    new_text, removed = _remove_marker_lines(current, marker)
    if removed:
        _write_crontab(new_text)
    print(f"removed={'true' if removed else 'false'}")
    print(f"marker={marker}")
    print(f"manifest={manifest_path}")
    return 0


def show_cron(config_path: str) -> int:
    if not command_exists("crontab"):
        raise OneRunError("crontab command not found on this system")
    manifest_path = Path(config_path).resolve()
    marker = _cron_marker(manifest_path)
    current = _read_crontab()
    lines = [line for line in current.splitlines() if marker in line]
    if not lines:
        print("installed=false")
        print(f"marker={marker}")
        print(f"manifest={manifest_path}")
        return 0
    print("installed=true")
    print(f"marker={marker}")
    print(f"manifest={manifest_path}")
    print(f"entry={lines[0]}")
    return 0



class OneRunError(Exception):
    """base one-run error"""


class ManifestError(OneRunError):
    """manifest parsing/validation errors"""


class TemplateError(OneRunError):
    """entrypoint templating errors"""


class BackendError(OneRunError):
    """backend runtime errors"""



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any

LEGACY_HINTS = {
    ("experiment", "random_seed"): "use experiment.seed",
    ("entrypoint", "script"): "use entrypoint.command",
    ("backend", "type"): "use backend.kind",
}


def warn_legacy_manifest(payload: Any, manifest_path: Path) -> None:
    if not isinstance(payload, dict):
        return

    for (section, key), hint in LEGACY_HINTS.items():
        section_value = payload.get(section)
        if isinstance(section_value, dict) and key in section_value:
            warnings.warn(
                (
                    f"legacy manifest key detected in {manifest_path}: "
                    f"{section}.{key} is deprecated; {hint}."
                ),
                RuntimeWarning,
                stacklevel=2,
            )



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import yaml

from one_run.constants import JSON_SCHEMA_VERSION
from one_run.event_manager import emit_event
from one_run.run_dir import RunPaths
from one_run.status_store import read_metadata, read_runtime, read_status, write_json_atomic


def append_event(
    paths: RunPaths,
    *,
    run_id: str,
    event: str,
    state: str | None = None,
    message: str | None = None,
) -> None:
    emit_event(
        paths,
        run_id=run_id,
        event=event,
        state=state,
        message=message,
    )


def build_summary_payload(paths: RunPaths, run_id: str) -> dict:
    metadata = read_metadata(paths.metadata) if paths.metadata.exists() else None
    status = read_status(paths.status).model_dump(mode="json") if paths.status.exists() else None
    runtime = read_runtime(paths.runtime).model_dump(mode="json") if paths.runtime.exists() else None
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "run_id": run_id,
        "metadata": metadata,
        "status": status,
        "runtime": runtime,
        "paths": {
            "run_dir": str(paths.run_dir),
            "config_snapshot": str(paths.config_snapshot),
            "metadata": str(paths.metadata),
            "status": str(paths.status),
            "runtime": str(paths.runtime),
            "events": str(paths.events),
            "stdout_log": str(paths.stdout_log),
            "stderr_log": str(paths.stderr_log),
            "artifacts_dir": str(paths.artifacts_dir),
            "metrics_dir": str(paths.metrics_dir),
        },
    }


def _extract_command(paths: RunPaths) -> str | None:
    if not paths.config_snapshot.exists():
        return None
    try:
        payload = yaml.safe_load(paths.config_snapshot.read_text(encoding="utf-8"))
    except yaml.YAMLError:
        return None
    if not isinstance(payload, dict):
        return None
    entrypoint = payload.get("entrypoint")
    if not isinstance(entrypoint, dict):
        return None
    command = entrypoint.get("command")
    return str(command) if isinstance(command, str) else None


def _artifact_paths(artifacts_dir: Path) -> list[str]:
    if not artifacts_dir.exists():
        return []
    files = [p for p in artifacts_dir.rglob("*") if p.is_file()]
    return sorted(str(p.relative_to(artifacts_dir)) for p in files)


def _flatten_scalars(value: Any, *, prefix: str = "") -> dict[str, str | int | float | bool]:
    if isinstance(value, dict):
        out: dict[str, str | int | float | bool] = {}
        for key, item in value.items():
            key_str = str(key)
            next_prefix = f"{prefix}.{key_str}" if prefix else key_str
            out.update(_flatten_scalars(item, prefix=next_prefix))
        return out
    if isinstance(value, (str, int, float, bool)):
        return {prefix: value} if prefix else {}
    return {}


def _metrics_key_values(metrics_dir: Path) -> dict[str, str | int | float | bool]:
    if not metrics_dir.exists():
        return {}
    json_files = sorted(p for p in metrics_dir.rglob("*.json") if p.is_file())
    if not json_files:
        return {}
    preferred = metrics_dir / "final.json"
    source = preferred if preferred in json_files else json_files[0]
    try:
        payload = json.loads(source.read_text(encoding="utf-8"))
    except json.JSONDecodeError:
        return {}
    return _flatten_scalars(payload)


def build_summary_min_payload(paths: RunPaths, run_id: str) -> dict:
    metadata = read_metadata(paths.metadata) if paths.metadata.exists() else {}
    status = read_status(paths.status).model_dump(mode="json") if paths.status.exists() else {}
    runtime = read_runtime(paths.runtime).model_dump(mode="json") if paths.runtime.exists() else {}
    return {
        "run_id": run_id,
        "state": status.get("state"),
        "exit_code": status.get("exit_code"),
        "start": status.get("started_at"),
        "end": status.get("ended_at"),
        "duration": runtime.get("duration_sec"),
        "command": _extract_command(paths),
        "tags": metadata.get("tags") or [],
        "artifacts": _artifact_paths(paths.artifacts_dir),
        "metrics_key_values": _metrics_key_values(paths.metrics_dir),
    }


def write_summary(paths: RunPaths, run_id: str) -> None:
    write_json_atomic(paths.summary, build_summary_payload(paths, run_id))
    write_json_atomic(paths.summary_min, build_summary_min_payload(paths, run_id))



from pathlib import Path

RUNS_DIR_NAME = "runs"
CONFIG_SNAPSHOT_NAME = "config.snapshot.yaml"
METADATA_NAME = "metadata.json"
STATUS_NAME = "status.json"
RUNTIME_NAME = "runtime.json"
SUMMARY_NAME = "summary.json"
SUMMARY_MIN_NAME = "summary_min.json"
EVENTS_NAME = "events.jsonl"
LOGS_DIR_NAME = "logs"
STDOUT_LOG_NAME = "stdout.log"
STDERR_LOG_NAME = "stderr.log"
ARTIFACTS_DIR_NAME = "artifacts"
METRICS_DIR_NAME = "metrics"
JOB_SCRIPT_NAME = "job.sbatch"
LOCAL_PID_NAME = "local.pid"
LOCAL_EXIT_CODE_NAME = "local.exit_code"

ALLOWED_PLACEHOLDERS = {"config", "run_dir", "seed", "project_root"}

DEFAULT_RUNS_DIR = Path(RUNS_DIR_NAME)
JSON_SCHEMA_VERSION = "one-run/v1"
AUTOMATION_ENV_VAR = "ONE_RUN_AUTOMATION"



from __future__ import annotations

import time
from datetime import datetime
from pathlib import Path
from typing import Any

from one_run.commands.events_common import event_matches, format_event_line, normalize_values
from one_run.commands.events_stream import consume_event_file, ensure_offset
from one_run.constants import DEFAULT_RUNS_DIR
from one_run.errors import OneRunError
from one_run.listeners import build_routes_from_manifest, load_handler, manifest_interval_sec
from one_run.utils.duration import parse_duration


def _iter_event_files(runs_root: Path, run_ids: list[str] | None = None) -> list[Path]:
    if not runs_root.exists():
        return []
    if run_ids:
        files: list[Path] = []
        for run_id in run_ids:
            path = runs_root / run_id / "events.jsonl"
            if not path.exists():
                raise OneRunError(f"event file not found: {path}")
            files.append(path)
        return files
    files: list[Path] = []
    for run_dir in runs_root.iterdir():
        if not run_dir.is_dir():
            continue
        events_path = run_dir / "events.jsonl"
        if events_path.exists():
            files.append(events_path)
    return sorted(files)


def listen_events(
    *,
    run_ids: list[str] | None = None,
    watch: bool = True,
    jsonl: bool = False,
    output: str = "table",
    interval_sec: float | None = None,
    events: list[str] | None = None,
    states: list[str] | None = None,
    tags: list[str] | None = None,
    handler: str | None = None,
    manifest_path: str | None = None,
    once: bool = False,
    history: bool = False,
    since: str | None = None,
) -> int:
    project_root = Path.cwd().resolve()
    runs_root = project_root / DEFAULT_RUNS_DIR
    allowed_events = normalize_values(events)
    allowed_states = normalize_values(states)
    required_tags = normalize_values(tags)
    tag_cache: dict[str, set[str]] = {}
    offsets: dict[Path, int] = {}
    callback = load_handler(handler)
    routes = build_routes_from_manifest(manifest_path)
    manifest_interval = manifest_interval_sec(manifest_path)
    poll_interval = interval_sec if interval_sec is not None else (manifest_interval if manifest_interval is not None else 1.0)
    render_jsonl = jsonl or output == "jsonl"
    if since and not history:
        raise OneRunError("--since requires --history")
    since_ts: datetime | None = None
    if since:
        try:
            since_ts = datetime.now() - parse_duration(since)
        except ValueError as exc:
            raise OneRunError(str(exc)) from exc

    if once:
        watch = False

    def _event_ts(payload: dict[str, Any]) -> datetime | None:
        raw = payload.get("timestamp")
        if not isinstance(raw, str):
            return None
        normalized = raw.replace("Z", "+00:00")
        try:
            parsed = datetime.fromisoformat(normalized)
        except ValueError:
            return None
        if parsed.tzinfo is not None:
            return parsed.replace(tzinfo=None)
        return parsed

    def _consume_new(path: Path) -> None:
        offset = ensure_offset(path, offsets=offsets, history=history)

        def _on_line(_: Path, text: str, payload: dict[str, Any] | None) -> None:
            if payload is None:
                return
            if since_ts is not None:
                event_ts = _event_ts(payload)
                if event_ts is None or event_ts < since_ts:
                    return
            if not event_matches(
                payload,
                allowed_events=allowed_events,
                allowed_states=allowed_states,
                required_tags=required_tags,
                project_root=project_root,
                tag_cache=tag_cache,
            ):
                return
            if routes:
                for route in routes:
                    if not event_matches(
                        payload,
                        allowed_events=route.allowed_events,
                        allowed_states=route.allowed_states,
                        required_tags=route.required_tags,
                        project_root=project_root,
                        tag_cache=tag_cache,
                    ):
                        continue
                    if render_jsonl:
                        print(text)
                    else:
                        print(f"[listener:{route.label}] {format_event_line(payload)}")
                    if route.callback is not None:
                        route.callback(payload)
            else:
                if render_jsonl:
                    print(text)
                else:
                    print(format_event_line(payload))
                if callback is not None:
                    callback(payload)

        offsets[path] = consume_event_file(path, offset=offset, on_line=_on_line)

    try:
        if watch:
            print(f"[listener] started watching at {datetime.now().strftime('%H:%M:%S')}")
        while True:
            for path in _iter_event_files(runs_root, run_ids):
                _consume_new(path)
            if not watch:
                return 0
            time.sleep(max(poll_interval, 0.1))
    except KeyboardInterrupt:
        print("[listener] stopped")
        return 0



from __future__ import annotations

import re
from datetime import timedelta

_DURATION_RE = re.compile(r"^(\d+)([smhd])$")


_MULTIPLIERS = {
    "s": 1,
    "m": 60,
    "h": 3600,
    "d": 86400,
}


def parse_duration(value: str) -> timedelta:
    match = _DURATION_RE.match(value.strip())
    if not match:
        raise ValueError("invalid duration format; use <int><s|m|h|d>, e.g. 30m or 7d")

    amount = int(match.group(1))
    unit = match.group(2)
    if amount <= 0:
        raise ValueError("duration must be > 0")

    return timedelta(seconds=amount * _MULTIPLIERS[unit])



from __future__ import annotations

import json
from datetime import datetime
from threading import Lock
from typing import Any, Callable, Protocol

from one_run.constants import JSON_SCHEMA_VERSION
from one_run.run_dir import RunPaths


class EventSink(Protocol):
    def emit(self, paths: RunPaths, payload: dict[str, Any]) -> None:
        ...


class JsonlEventSink:
    def emit(self, paths: RunPaths, payload: dict[str, Any]) -> None:
        paths.events.parent.mkdir(parents=True, exist_ok=True)
        with paths.events.open("a", encoding="utf-8") as f:
            f.write(json.dumps(payload))
            f.write("\n")


class EventManager:
    def __init__(self, sinks: list[EventSink]) -> None:
        self._sinks = sinks

    def emit(self, paths: RunPaths, payload: dict[str, Any]) -> None:
        for sink in self._sinks:
            sink.emit(paths, payload)


EventSinkFactory = Callable[[], EventSink]

_sink_factories: list[EventSinkFactory] = []
_sink_factories_lock = Lock()


def register_event_sink_factory(factory: EventSinkFactory) -> Callable[[], None]:
    with _sink_factories_lock:
        _sink_factories.append(factory)

    def _unregister() -> None:
        with _sink_factories_lock:
            try:
                _sink_factories.remove(factory)
            except ValueError:
                return

    return _unregister


def build_event_payload(
    *,
    run_id: str,
    event: str,
    state: str | None = None,
    message: str | None = None,
    extra: dict[str, Any] | None = None,
) -> dict[str, Any]:
    payload: dict[str, Any] = {
        "schema_version": JSON_SCHEMA_VERSION,
        "timestamp": datetime.now().isoformat(),
        "run_id": run_id,
        "event": event,
    }
    if state is not None:
        payload["state"] = state
    if message is not None:
        payload["message"] = message
    if extra:
        payload.update(extra)
    return payload


def default_event_manager() -> EventManager:
    with _sink_factories_lock:
        extra = [factory() for factory in _sink_factories]
    return EventManager([JsonlEventSink(), *extra])


def emit_event(
    paths: RunPaths,
    *,
    run_id: str,
    event: str,
    state: str | None = None,
    message: str | None = None,
    extra: dict[str, Any] | None = None,
) -> None:
    payload = build_event_payload(
        run_id=run_id,
        event=event,
        state=state,
        message=message,
        extra=extra,
    )
    default_event_manager().emit(paths, payload)



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from one_run.constants import DEFAULT_RUNS_DIR
from one_run.utils.timefmt import iso_short


def normalize_values(values: list[str] | None) -> set[str]:
    if not values:
        return set()
    return {str(v).strip() for v in values if str(v).strip()}


def format_event_line(payload: dict[str, Any]) -> str:
    timestamp = iso_short(payload.get("timestamp"))
    run_id = payload.get("run_id", "?")
    event = payload.get("event", "?")
    state = payload.get("state", "-")
    message = payload.get("message", "-")
    return f"[{run_id}] timestamp={timestamp} event={event} state={state} message={message}"


def load_run_tags(project_root: Path, run_id: str, cache: dict[str, set[str]]) -> set[str]:
    cached = cache.get(run_id)
    if cached is not None:
        return cached
    metadata_path = project_root / DEFAULT_RUNS_DIR / run_id / "metadata.json"
    if not metadata_path.exists():
        cache[run_id] = set()
        return set()
    try:
        payload = json.loads(metadata_path.read_text(encoding="utf-8"))
    except json.JSONDecodeError:
        cache[run_id] = set()
        return set()
    raw_tags = payload.get("tags")
    if not isinstance(raw_tags, list):
        cache[run_id] = set()
        return set()
    tags = {str(tag).strip() for tag in raw_tags if str(tag).strip()}
    cache[run_id] = tags
    return tags


def event_matches(
    payload: dict[str, Any],
    *,
    allowed_events: set[str],
    allowed_states: set[str],
    required_tags: set[str],
    project_root: Path,
    tag_cache: dict[str, set[str]],
) -> bool:
    event_name = str(payload.get("event", "")).strip()
    state = str(payload.get("state", "")).strip()
    if allowed_events and event_name not in allowed_events:
        return False
    if allowed_states and state not in allowed_states:
        return False
    if not required_tags:
        return True
    run_id = str(payload.get("run_id", "")).strip()
    if not run_id:
        return False
    tags = load_run_tags(project_root, run_id, tag_cache)
    return required_tags.issubset(tags)



from __future__ import annotations

from pathlib import Path

import yaml

from one_run.errors import OneRunError
from one_run.models import ListenerConfig
from one_run.listeners.routes import ListenerRoute, route_from_listener


def _load_manifest_payload(manifest_path: str | None) -> dict:
    if not manifest_path:
        return {}
    path = Path(manifest_path)
    if not path.exists():
        raise OneRunError(f"manifest not found: {path}")
    try:
        payload = yaml.safe_load(path.read_text(encoding="utf-8"))
    except yaml.YAMLError as exc:
        raise OneRunError(f"invalid yaml in {path}: {exc}") from exc
    if not isinstance(payload, dict):
        raise OneRunError(f"invalid listener manifest: expected mapping in {path}")
    return payload


def build_routes_from_manifest(manifest_path: str | None) -> list[ListenerRoute]:
    payload = _load_manifest_payload(manifest_path)
    if not payload:
        return []
    raw_listeners = payload.get("listeners")
    if raw_listeners is None:
        raise OneRunError(f"listeners section not found in {manifest_path}")
    if not isinstance(raw_listeners, list):
        raise OneRunError(f"listeners must be a list in {manifest_path}")
    listeners = [ListenerConfig.model_validate(item) for item in raw_listeners]
    return [route_from_listener(listener, idx=idx) for idx, listener in enumerate(listeners)]


def manifest_interval_sec(manifest_path: str | None) -> float | None:
    payload = _load_manifest_payload(manifest_path)
    if not payload:
        return None
    raw = payload.get("interval_sec")
    if raw is None:
        return None
    try:
        value = float(raw)
    except (TypeError, ValueError) as exc:
        raise OneRunError(f"interval_sec must be a positive number in {manifest_path}") from exc
    if value <= 0:
        raise OneRunError(f"interval_sec must be > 0 in {manifest_path}")
    return value



from __future__ import annotations

import json
import os
import time
from datetime import datetime
from pathlib import Path

from one_run.backends.base import RunContext
from one_run.backends.local import LocalBackend
from one_run.backends.slurm import SlurmBackend
from one_run.commands.common import load_run_paths
from one_run.utils.duration import parse_duration
from one_run.errors import OneRunError
from one_run.manifest import load_manifest
from one_run.models import Manifest, RunMetadata, RunRuntime, RunStatus
from one_run.payloads import status_result_payload
from one_run.run_dir import RunPaths
from one_run.runtime import slurm_runtime_snapshot
from one_run.utils.shell import run_capture
from one_run.status_store import read_metadata, read_runtime, read_status, write_metadata, write_runtime, write_status
from one_run.utils.timefmt import iso_short
from one_run.tracking import append_event, write_summary


_TERMINAL_STATES = {"success", "failed"}


def _retry_category(raw_state: str | None) -> str:
    state = (raw_state or "").upper()
    if "TIMEOUT" in state:
        return "timeout"
    if "CANCELLED" in state or "PREEMPTED" in state:
        return "cancelled"
    return "failed"


def _resubmit_slurm_job(paths: RunPaths, project_root: Path) -> str:
    script_path = paths.run_dir / "job.sbatch"
    if not script_path.exists():
        raise OneRunError(f"cannot retry slurm run: missing {script_path}")
    proc = run_capture(["sbatch", str(script_path)], cwd=project_root)
    if proc.returncode != 0:
        msg = (proc.stderr or proc.stdout).strip() or "sbatch failed"
        raise OneRunError(f"failed to retry slurm job: {msg}")
    job_id = SlurmBackend.parse_sbatch_job_id(proc.stdout or "")
    if not job_id:
        raise OneRunError(f"failed to parse slurm job id from retry output: {(proc.stdout or '').strip()}")
    return job_id


def _maybe_retry_slurm(
    run_id: str,
    project_root: Path,
    paths: RunPaths,
    metadata: RunMetadata,
    status: RunStatus,
    query_runtime: dict | None,
) -> RunStatus:
    try:
        manifest = load_manifest(paths.config_snapshot)
    except Exception:
        return status

    slurm_cfg = manifest.backend.slurm
    if not slurm_cfg:
        return status

    max_attempts = slurm_cfg.retry_max_attempts or 0
    if max_attempts <= 0:
        return status

    on_states = set(slurm_cfg.retry_on_states or ["failed", "timeout", "cancelled"])
    raw_state = (query_runtime or {}).get("slurm_state")
    category = _retry_category(raw_state)
    if category not in on_states:
        return status

    current_attempt = metadata.slurm_attempt or 1
    if current_attempt > max_attempts:
        return status

    backoff_sec = slurm_cfg.retry_backoff_sec or 0
    if backoff_sec > 0 and status.ended_at is not None:
        elapsed = (datetime.now() - status.ended_at).total_seconds()
        if elapsed < backoff_sec:
            remaining = int(backoff_sec - elapsed)
            pending = status.model_copy(
                update={"message": f"retry pending in {remaining}s (attempt {current_attempt + 1}/{max_attempts + 1})"}
            )
            if pending != status:
                write_status(paths.status, pending)
                append_event(paths, run_id=run_id, event="retry_pending", state=pending.state, message=pending.message)
                write_summary(paths, run_id)
            return pending

    new_job_id = _resubmit_slurm_job(paths, project_root)
    next_attempt = current_attempt + 1
    updated_metadata = metadata.model_copy(update={"slurm_job_id": new_job_id, "slurm_attempt": next_attempt})
    write_metadata(paths.metadata, updated_metadata.model_dump(mode="json"))

    retried = status.model_copy(
        update={
            "state": "pending",
            "started_at": datetime.now(),
            "ended_at": None,
            "exit_code": None,
            "message": f"retrying attempt {next_attempt}/{max_attempts + 1}",
        }
    )
    write_status(paths.status, retried)
    append_event(paths, run_id=run_id, event="retry_submitted", state="pending", message=retried.message)
    write_summary(paths, run_id)
    return retried


def _refresh_slurm_if_possible(run_id: str, project_root: Path, status: RunStatus, metadata: RunMetadata) -> RunStatus:
    if metadata.backend_type != "slurm":
        return status

    dummy_manifest = Manifest.model_validate(
        {
            "experiment": {"name": run_id, "seed": 0},
            "entrypoint": {"command": "echo noop"},
            "environment": {"kind": "local"},
            "backend": {"kind": "slurm"},
        }
    )

    paths = load_run_paths(project_root, run_id)
    run_ctx = RunContext(
        run_id=run_id,
        manifest_path=paths.config_snapshot,
        manifest=dummy_manifest,
        paths=paths,
        project_root=project_root,
        metadata=metadata,
    )

    query = SlurmBackend().query(run_ctx, status)
    if query is None:
        return status

    ended_at = status.ended_at
    if query.state in {"success", "failed"} and ended_at is None:
        ended_at = datetime.now()

    updated = status.model_copy(update={"state": query.state, "message": query.message, "ended_at": ended_at})
    if updated != status:
        write_status(paths.status, updated)
        append_event(paths, run_id=run_id, event="state_changed", state=updated.state, message=updated.message)
        write_summary(paths, run_id)
    if query.runtime:
        runtime = slurm_runtime_snapshot(
            elapsed=query.runtime.get("slurm_elapsed"),
            max_rss=query.runtime.get("slurm_max_rss"),
            alloc_cpus=query.runtime.get("slurm_alloc_cpus"),
        )
        write_runtime(paths.runtime, RunRuntime.model_validate(runtime.model_dump(mode="json")))
        write_summary(paths, run_id)
    if updated.state == "failed":
        updated = _maybe_retry_slurm(
            run_id=run_id,
            project_root=project_root,
            paths=paths,
            metadata=metadata,
            status=updated,
            query_runtime=query.runtime,
        )
    return updated


def _refresh_local_if_possible(paths: RunPaths, status: RunStatus) -> RunStatus:
    if status.state not in {"pending", "running"}:
        return status

    if paths.local_exit_code.exists():
        try:
            exit_code = int(paths.local_exit_code.read_text(encoding="utf-8").strip())
        except ValueError:
            exit_code = 1
        ended_at = status.ended_at or datetime.fromtimestamp(paths.local_exit_code.stat().st_mtime)
        updated = status.model_copy(
            update={
                "state": "success" if exit_code == 0 else "failed",
                "exit_code": exit_code,
                "ended_at": ended_at,
                "message": None if exit_code == 0 else f"command exited with code {exit_code}",
            }
        )
        write_status(paths.status, updated)
        append_event(paths, run_id=paths.run_dir.name, event="state_changed", state=updated.state, message=updated.message)
        write_summary(paths, paths.run_dir.name)
        return updated

    if paths.local_pid.exists():
        try:
            pid = int(paths.local_pid.read_text(encoding="utf-8").strip())
            os.kill(pid, 0)
        except ProcessLookupError:
            updated = status.model_copy(
                update={
                    "state": "failed",
                    "ended_at": status.ended_at or datetime.now(),
                    "message": "process ended without exit code file",
                }
            )
            write_status(paths.status, updated)
            append_event(paths, run_id=paths.run_dir.name, event="state_changed", state=updated.state, message=updated.message)
            write_summary(paths, paths.run_dir.name)
            return updated
        except (ValueError, OSError):
            pass

    return status


def show_status(
    run_id: str,
    *,
    json_output: bool = False,
    long: bool = False,
    wait: bool = False,
    timeout: str | None = None,
    interval_sec: float = 2.0,
) -> int:
    project_root = Path.cwd().resolve()
    paths = load_run_paths(project_root, run_id)
    timeout_sec = None
    if timeout:
        try:
            timeout_sec = parse_duration(timeout).total_seconds()
        except ValueError as exc:
            raise OneRunError(str(exc)) from exc

    started_wait_at = datetime.now()
    timed_out = False

    while True:
        metadata = RunMetadata.model_validate(read_metadata(paths.metadata))
        status = read_status(paths.status)

        if metadata.backend_type == "slurm":
            status = _refresh_slurm_if_possible(run_id, project_root, status, metadata)
        else:
            status = _refresh_local_if_possible(paths, status)
            LocalBackend().query(
                RunContext(
                    run_id=run_id,
                    manifest_path=paths.config_snapshot,
                    manifest=Manifest.model_validate(
                        {
                            "experiment": {"name": run_id, "seed": 0},
                            "entrypoint": {"command": "echo noop"},
                            "environment": {"kind": "local"},
                            "backend": {"kind": "local"},
                        }
                    ),
                    paths=paths,
                    project_root=project_root,
                    metadata=metadata,
                ),
                status,
            )

        if not wait or status.state in _TERMINAL_STATES:
            break

        if timeout_sec is not None:
            elapsed = (datetime.now() - started_wait_at).total_seconds()
            if elapsed >= timeout_sec:
                timed_out = True
                break

        time.sleep(max(interval_sec, 0.1))

    runtime = read_runtime(paths.runtime) if paths.runtime.exists() else None
    write_summary(paths, run_id)
    if json_output:
        payload = status_result_payload(
            run_id=run_id,
            wait=wait,
            timed_out=timed_out,
            backend=metadata.backend_type,
            metadata=metadata,
            status=status,
            runtime=runtime,
        )
        print(json.dumps(payload))
        return 1 if timed_out else 0

    if long:
        print(f"state={status.state}")
        print(f"backend={metadata.backend_type}")
        print(f"timestamp={iso_short(metadata.timestamp)}")
        print(f"started_at={iso_short(status.started_at)}")
        print(f"ended_at={iso_short(status.ended_at)}")
        print(f"exit_code={status.exit_code}")
        if status.message:
            print(f"message={status.message}")
        print(f"hostname={metadata.hostname}")
        print(f"cwd={metadata.cwd}")
        print(f"git_commit={metadata.git_commit}")
        print(f"git_dirty={metadata.git_dirty}")
        if metadata.docker_image_digest:
            print(f"docker_image_digest={metadata.docker_image_digest}")
        if runtime:
            if runtime.duration_sec is not None:
                print(f"duration_sec={runtime.duration_sec}")
            if runtime.peak_rss_mb is not None:
                print(f"peak_rss_mb={runtime.peak_rss_mb}")
        print(f"stdout_log={paths.stdout_log}")
        print(f"stderr_log={paths.stderr_log}")
        print(f"artifacts_dir={paths.artifacts_dir}")
        return 0

    print(f"run_id={run_id}")
    print(f"state={status.state}")
    print(f"backend={metadata.backend_type}")
    if metadata.slurm_job_id:
        print(f"slurm_job_id={metadata.slurm_job_id}")
    if status.exit_code is not None:
        print(f"exit_code={status.exit_code}")
    if status.message:
        print(f"message={status.message}")
    if wait:
        print(f"timed_out={str(timed_out).lower()}")
    return 1 if timed_out else 0



from __future__ import annotations

import importlib
import json
from dataclasses import dataclass
from typing import Any, Callable
from urllib import request
from urllib.error import URLError

from one_run.errors import OneRunError
from one_run.models import ListenerConfig
from one_run.commands.events_common import normalize_values


def load_handler(handler: str | None) -> Callable[[dict[str, Any]], None] | None:
    if not handler:
        return None
    if ":" not in handler:
        raise OneRunError("handler must be in format module:function")
    module_name, func_name = handler.split(":", 1)
    try:
        mod = importlib.import_module(module_name)
    except Exception as exc:
        raise OneRunError(f"failed to import handler module: {module_name}") from exc
    fn = getattr(mod, func_name, None)
    if not callable(fn):
        raise OneRunError(f"handler is not callable: {handler}")
    return fn


@dataclass
class ListenerRoute:
    label: str
    kind: str
    action: str | None
    webhook_url: str | None
    allowed_events: set[str]
    allowed_states: set[str]
    required_tags: set[str]
    callback: Callable[[dict[str, Any]], None] | None


def _post_json(url: str, payload: dict[str, Any], *, timeout_sec: float) -> None:
    data = json.dumps(payload).encode("utf-8")
    req = request.Request(url, data=data, method="POST", headers={"content-type": "application/json"})
    with request.urlopen(req, timeout=timeout_sec) as resp:
        code = getattr(resp, "status", 200)
        if int(code) >= 400:
            raise OneRunError(f"webhook returned status {code}")


def _make_webhook_callback(
    *,
    webhook_url: str,
    timeout_sec: float,
    action: str | None,
) -> Callable[[dict[str, Any]], None]:
    def _callback(event_payload: dict[str, Any]) -> None:
        out = {
            "kind": "webhook",
            "action": action or "event",
            "event": event_payload,
        }
        try:
            _post_json(webhook_url, out, timeout_sec=timeout_sec)
        except (URLError, TimeoutError, OneRunError):
            return

    return _callback


def route_from_listener(listener: ListenerConfig, *, idx: int) -> ListenerRoute:
    label = listener.name or f"{listener.kind}-{idx + 1}"
    callback: Callable[[dict[str, Any]], None] | None
    if listener.kind == "local":
        callback = load_handler(listener.handler) if listener.handler else None
    else:
        callback = _make_webhook_callback(
            webhook_url=listener.webhook_url or "",
            timeout_sec=listener.timeout_sec or 3.0,
            action=listener.action,
        )
    return ListenerRoute(
        label=label,
        kind=listener.kind,
        action=listener.action,
        webhook_url=listener.webhook_url,
        allowed_events=normalize_values(listener.events),
        allowed_states=normalize_values(listener.states),
        required_tags=normalize_values(listener.tags),
        callback=callback,
    )



from __future__ import annotations

import json
from pathlib import Path
from tempfile import NamedTemporaryFile

from one_run.models import RunRuntime, RunStatus


def write_json_atomic(path: Path, data: dict) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with NamedTemporaryFile("w", encoding="utf-8", delete=False, dir=path.parent) as tmp:
        json.dump(data, tmp, indent=2, default=str)
        tmp.write("\n")
        tmp_path = Path(tmp.name)
    tmp_path.replace(path)


def write_status(path: Path, status: RunStatus) -> None:
    write_json_atomic(path, status.model_dump(mode="json"))


def read_status(path: Path) -> RunStatus:
    payload = json.loads(path.read_text(encoding="utf-8"))
    return RunStatus.model_validate(payload)


def write_metadata(path: Path, metadata: dict) -> None:
    write_json_atomic(path, metadata)


def read_metadata(path: Path) -> dict:
    return json.loads(path.read_text(encoding="utf-8"))


def write_runtime(path: Path, runtime: RunRuntime) -> None:
    write_json_atomic(path, runtime.model_dump(mode="json"))


def read_runtime(path: Path) -> RunRuntime:
    payload = json.loads(path.read_text(encoding="utf-8"))
    return RunRuntime.model_validate(payload)



from __future__ import annotations

import json
import shutil
from datetime import datetime
from pathlib import Path

from one_run.commands.cancel_cmd import _cancel_local, _cancel_slurm
from one_run.commands.common import load_run_paths
from one_run.utils.duration import parse_duration
from one_run.errors import OneRunError
from one_run.models import RunMetadata
from one_run.payloads import rm_cleanup_result_payload, rm_result_payload
from one_run.status_store import read_metadata, read_status


def remove_run(run_id: str, *, force: bool = False, json_output: bool = False) -> int:
    project_root = Path.cwd().resolve()
    paths = load_run_paths(project_root, run_id)
    metadata = RunMetadata.model_validate(read_metadata(paths.metadata))
    status = read_status(paths.status)

    state_before = status.state
    cancelled = False
    state_after = status.state

    if status.state not in {"success", "failed"}:
        if metadata.backend_type == "slurm":
            updated = _cancel_slurm(metadata, status, paths, force=force)
        else:
            updated = _cancel_local(paths, status, force=force)
        cancelled = True
        state_after = updated.state

    shutil.rmtree(paths.run_dir)

    if json_output:
        print(
            json.dumps(
                rm_result_payload(
                    run_id=run_id,
                    removed=True,
                    cancelled=cancelled,
                    force=force,
                    state_before=state_before,
                    state_after=state_after,
                )
            )
        )
    else:
        print(f"run_id={run_id}")
        print("removed=true")
        print(f"cancelled={str(cancelled).lower()}")
        print(f"force={str(force).lower()}")
    return 0


def _reference_time(status, metadata: dict) -> datetime | None:
    if status.ended_at:
        return status.ended_at
    if status.started_at:
        return status.started_at
    ts = metadata.get("timestamp")
    if not ts:
        return None
    try:
        return datetime.fromisoformat(str(ts))
    except ValueError:
        return None


def remove_runs(
    *,
    state: str | None = None,
    older_than: str | None = None,
    dry_run: bool = False,
    json_output: bool = False,
) -> int:
    project_root = Path.cwd().resolve()
    runs_root = project_root / "runs"
    if not runs_root.exists():
        if json_output:
            print(json.dumps(rm_cleanup_result_payload(items=[], count=0, dry_run=dry_run)))
        else:
            print("no runs found")
        return 0

    cutoff = None
    if older_than:
        try:
            cutoff = datetime.now() - parse_duration(older_than)
        except ValueError as exc:
            raise OneRunError(str(exc)) from exc

    selected: list[Path] = []
    for run_dir in sorted([p for p in runs_root.iterdir() if p.is_dir()], reverse=True):
        metadata_path = run_dir / "metadata.json"
        status_path = run_dir / "status.json"
        if not metadata_path.exists() or not status_path.exists():
            continue
        metadata = read_metadata(metadata_path)
        status_obj = read_status(status_path)
        if state and status_obj.state != state:
            continue
        ref_ts = _reference_time(status_obj, metadata)
        if cutoff and ref_ts and ref_ts >= cutoff:
            continue
        selected.append(run_dir)

    if not selected:
        if json_output:
            print(json.dumps(rm_cleanup_result_payload(items=[], count=0, dry_run=dry_run)))
        else:
            print("no runs matched filter")
        return 0

    items: list[dict[str, str]] = []
    for run_dir in selected:
        if dry_run:
            items.append({"run_id": run_dir.name, "action": "would_delete"})
            if not json_output:
                print(f"would_delete={run_dir.name}")
        else:
            shutil.rmtree(run_dir)
            items.append({"run_id": run_dir.name, "action": "deleted"})
            if not json_output:
                print(f"deleted={run_dir.name}")

    if json_output:
        print(json.dumps(rm_cleanup_result_payload(items=items, count=len(selected), dry_run=dry_run)))
    else:
        print(f"count={len(selected)}")
        print(f"dry_run={str(dry_run).lower()}")
    return 0



from one_run.utils.duration import parse_duration
from one_run.utils.shell import command_exists, run_capture
from one_run.utils.templating import render_command, validate_placeholders
from one_run.utils.terminal import colorize, render_table, should_use_color
from one_run.utils.timefmt import iso_short

__all__ = [
    "command_exists",
    "run_capture",
    "parse_duration",
    "validate_placeholders",
    "render_command",
    "iso_short",
    "should_use_color",
    "colorize",
    "render_table",
]



from __future__ import annotations

from datetime import datetime


def make_run_id(experiment_name: str, now: datetime | None = None) -> str:
    _ = experiment_name
    ts = (now or datetime.now()).strftime("%m%d%H%M%S")
    return ts



from __future__ import annotations

import json
from pathlib import Path

from one_run.manifest import load_manifest
from one_run.payloads import validate_result_payload
from one_run.utils.templating import validate_placeholders


def validate_manifest(config_path: str, *, json_output: bool = False) -> int:
    manifest_path = Path(config_path).resolve()
    manifest = load_manifest(manifest_path)
    validate_placeholders(manifest.entrypoint.command)

    if json_output:
        print(json.dumps(validate_result_payload(
            manifest=str(manifest_path),
            valid=True,
            backend=manifest.backend.kind,
            environment=manifest.environment.kind,
        )))
        return 0

    print(f"manifest={manifest_path}")
    print("valid=true")
    print(f"backend={manifest.backend.kind}")
    print(f"environment={manifest.environment.kind}")
    return 0



from __future__ import annotations

import json
import os
from datetime import datetime
from pathlib import Path
from typing import Any

from one_run.constants import DEFAULT_RUNS_DIR
from one_run.models import RunStatus
from one_run.payloads import ls_result_payload
from one_run.run_dir import resolve_run_paths
from one_run.utils.shell import run_capture
from one_run.status_store import read_metadata, read_status, write_status
from one_run.utils.terminal import colorize, render_table, should_use_color
from one_run.utils.timefmt import iso_short
from one_run.tracking import append_event, write_summary


def _fmt(value: Any) -> str:
    return "-" if value is None else str(value)


def _format_row(row: dict[str, Any], *, long: bool) -> dict[str, str]:
    if long:
        return {
            "run_id": row["run_id"],
            "experiment": row["experiment"],
            "state": row["state"],
            "backend": row["backend"],
            "timestamp": row["timestamp"],
            "tags": ",".join(row["tags"]) or "-",
            "started_at": row["started_at"],
            "ended_at": row["ended_at"],
            "exit_code": row["exit_code"],
            "message": row["message"],
        }
    return {
        "run_id": row["run_id"],
        "experiment": row["experiment"],
        "state": row["state"],
        "backend": row["backend"],
        "timestamp": row["timestamp"],
        "tags": ",".join(row["tags"]) or "-",
    }


def _normalize_tags(raw_tags: Any) -> list[str]:
    if not isinstance(raw_tags, list):
        return []
    return sorted({str(tag).strip() for tag in raw_tags if str(tag).strip()})


def _matches_tags(run_tags: list[str], required_tags: list[str] | None) -> bool:
    if not required_tags:
        return True
    return set(required_tags).issubset(set(run_tags))


def _parse_timestamp_sort_key(value: Any) -> float:
    if not value:
        return float("-inf")
    text = str(value).strip()
    if not text:
        return float("-inf")
    if text.endswith("Z"):
        text = text[:-1] + "+00:00"
    try:
        return datetime.fromisoformat(text).timestamp()
    except ValueError:
        return float("-inf")


def _is_pid_alive(pid: int) -> bool:
    try:
        os.kill(pid, 0)
        return True
    except ProcessLookupError:
        return False
    except PermissionError:
        return True


def _is_pid_zombie(pid: int) -> bool:
    try:
        proc = run_capture(["ps", "-o", "stat=", "-p", str(pid)])
    except OSError:
        return False
    if proc.returncode != 0:
        return False
    stat = (proc.stdout or "").strip()
    return "Z" in stat


def _refresh_local_status_if_possible(run_dir: Path, status: RunStatus) -> RunStatus:
    if status.state not in {"pending", "running"}:
        return status

    local_exit_code = run_dir / "local.exit_code"
    local_pid = run_dir / "local.pid"
    status_path = run_dir / "status.json"

    if local_exit_code.exists():
        try:
            exit_code = int(local_exit_code.read_text(encoding="utf-8").strip())
        except ValueError:
            exit_code = 1
        ended_at = status.ended_at or datetime.fromtimestamp(local_exit_code.stat().st_mtime)
        updated = status.model_copy(
            update={
                "state": "success" if exit_code == 0 else "failed",
                "exit_code": exit_code,
                "ended_at": ended_at,
                "message": None if exit_code == 0 else f"command exited with code {exit_code}",
            }
        )
        if updated != status:
            write_status(status_path, updated)
            resolved = resolve_run_paths(run_dir)
            append_event(resolved, run_id=run_dir.name, event="state_changed", state=updated.state, message=updated.message)
            write_summary(resolved, run_dir.name)
        return updated

    if local_pid.exists():
        try:
            pid = int(local_pid.read_text(encoding="utf-8").strip())
        except ValueError:
            pid = -1
        if pid > 0 and (_is_pid_zombie(pid) or not _is_pid_alive(pid)):
            updated = status.model_copy(
                update={
                    "state": "failed",
                    "ended_at": status.ended_at or datetime.now(),
                    "message": "process ended without exit code file",
                }
            )
            if updated != status:
                write_status(status_path, updated)
                resolved = resolve_run_paths(run_dir)
                append_event(resolved, run_id=run_dir.name, event="state_changed", state=updated.state, message=updated.message)
                write_summary(resolved, run_dir.name)
            return updated

    return status


def list_runs(
    *,
    state_filter: str | None = None,
    long: bool = False,
    tags: list[str] | None = None,
    group_by_tag: bool = False,
    json_output: bool = False,
) -> int:
    project_root = Path.cwd().resolve()
    runs_root = project_root / DEFAULT_RUNS_DIR
    if not runs_root.exists():
        if json_output:
            print(json.dumps(ls_result_payload(items=[])))
        else:
            print("no runs found")
        return 0

    run_dirs = sorted([p for p in runs_root.iterdir() if p.is_dir()], reverse=True)
    if not run_dirs:
        if json_output:
            print(json.dumps(ls_result_payload(items=[])))
        else:
            print("no runs found")
        return 0

    selected_tags = sorted({tag.strip() for tag in (tags or []) if tag and tag.strip()})
    use_color = should_use_color()
    rows_with_sort: list[tuple[float, dict[str, Any]]] = []
    matched = 0
    for run_dir in run_dirs:
        metadata_path = run_dir / "metadata.json"
        status_path = run_dir / "status.json"
        if not metadata_path.exists() or not status_path.exists():
            continue
        metadata = read_metadata(metadata_path)
        status = read_status(status_path)
        if metadata.get("backend_type") == "local":
            status = _refresh_local_status_if_possible(run_dir, status)
        if state_filter and status.state != state_filter:
            continue
        run_tags = _normalize_tags(metadata.get("tags"))
        if not _matches_tags(run_tags, selected_tags):
            continue

        matched += 1
        rows_with_sort.append(
            (
                _parse_timestamp_sort_key(metadata.get("timestamp")),
                {
                    "run_id": run_dir.name,
                    "state": status.state,
                    "backend": metadata.get("backend_type", "?"),
                    "timestamp": iso_short(metadata.get("timestamp")),
                    "started_at": iso_short(status.started_at),
                    "ended_at": iso_short(status.ended_at),
                    "exit_code": _fmt(status.exit_code),
                    "message": _fmt(status.message),
                    "experiment": metadata.get("experiment_name", "?"),
                    "experiment_color": metadata.get("experiment_color"),
                    "tags": run_tags,
                },
            )
        )
    rows_with_sort.sort(key=lambda item: item[0], reverse=True)
    rows = [row for _, row in rows_with_sort]

    if matched == 0:
        if json_output:
            print(json.dumps(ls_result_payload(items=[])))
        else:
            print("no runs matched filter")
        return 0

    if json_output:
        print(json.dumps(ls_result_payload(items=rows)))
        return 0

    columns = (
        ["run_id", "experiment", "state", "backend", "timestamp", "tags", "started_at", "ended_at", "exit_code", "message"]
        if long
        else ["run_id", "experiment", "state", "backend", "timestamp", "tags"]
    )

    if group_by_tag:
        groups: dict[str, list[dict[str, Any]]] = {}
        for row in rows:
            row_tags = row["tags"] if row["tags"] else ["untagged"]
            for tag_name in row_tags:
                groups.setdefault(tag_name, []).append(row)

        for tag_name in sorted(groups):
            print(f"tag={tag_name}")
            display_rows = [_format_row(row, long=long) for row in groups[tag_name]]
            lines = render_table(columns, display_rows)
            print(lines[0])
            for idx, line in enumerate(lines):
                if idx == 0:
                    continue
                row = groups[tag_name][idx - 1]
                print(colorize(line, row["experiment_color"], enabled=use_color))
    else:
        display_rows = [_format_row(row, long=long) for row in rows]
        lines = render_table(columns, display_rows)
        print(lines[0])
        for idx, line in enumerate(lines[1:]):
            print(colorize(line, rows[idx]["experiment_color"], enabled=use_color))
    return 0



from __future__ import annotations

from pathlib import Path

from one_run.constants import DEFAULT_RUNS_DIR
from one_run.errors import BackendError, OneRunError
from one_run.run_dir import RunPaths, allocate_run_paths, resolve_run_paths


def alloc_run_dir(project_root: Path, experiment_name: str) -> tuple[str, RunPaths]:
    runs_root = project_root / DEFAULT_RUNS_DIR
    runs_root.mkdir(parents=True, exist_ok=True)
    try:
        return allocate_run_paths(runs_root, experiment_name)
    except FileExistsError as exc:
        raise BackendError(str(exc)) from exc


def load_run_paths(project_root: Path, run_id: str) -> RunPaths:
    run_dir = project_root / DEFAULT_RUNS_DIR / run_id
    if not run_dir.exists():
        raise OneRunError(f"run not found: {run_id}")
    return resolve_run_paths(run_dir)



from __future__ import annotations

import json
import os
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import Any

from one_run.constants import DEFAULT_RUNS_DIR
from one_run.models import RunStatus
from one_run.payloads import ps_result_payload
from one_run.run_dir import resolve_run_paths
from one_run.utils.shell import run_capture
from one_run.status_store import read_metadata, read_status, write_status
from one_run.utils.terminal import colorize, render_table, should_use_color
from one_run.utils.timefmt import iso_short
from one_run.tracking import append_event, write_summary


def _normalize_tags(raw_tags: Any) -> list[str]:
    if not isinstance(raw_tags, list):
        return []
    return sorted({str(tag).strip() for tag in raw_tags if str(tag).strip()})


def _matches_tags(run_tags: list[str], required_tags: list[str] | None) -> bool:
    if not required_tags:
        return True
    return set(required_tags).issubset(set(run_tags))


def _is_pid_alive(pid: int) -> bool:
    try:
        os.kill(pid, 0)
        return True
    except ProcessLookupError:
        return False
    except PermissionError:
        return True


def _is_pid_zombie(pid: int) -> bool:
    try:
        proc = run_capture(["ps", "-o", "stat=", "-p", str(pid)])
    except OSError:
        return False
    if proc.returncode != 0:
        return False
    stat = (proc.stdout or "").strip()
    return "Z" in stat


def _refresh_local_status_if_possible(run_dir: Path, status: RunStatus) -> RunStatus:
    if status.state not in {"pending", "running"}:
        return status

    local_exit_code = run_dir / "local.exit_code"
    local_pid = run_dir / "local.pid"
    status_path = run_dir / "status.json"

    if local_exit_code.exists():
        try:
            exit_code = int(local_exit_code.read_text(encoding="utf-8").strip())
        except ValueError:
            exit_code = 1
        ended_at = status.ended_at or datetime.fromtimestamp(local_exit_code.stat().st_mtime)
        updated = status.model_copy(
            update={
                "state": "success" if exit_code == 0 else "failed",
                "exit_code": exit_code,
                "ended_at": ended_at,
                "message": None if exit_code == 0 else f"command exited with code {exit_code}",
            }
        )
        if updated != status:
            write_status(status_path, updated)
            resolved = resolve_run_paths(run_dir)
            append_event(resolved, run_id=run_dir.name, event="state_changed", state=updated.state, message=updated.message)
            write_summary(resolved, run_dir.name)
        return updated

    if local_pid.exists():
        try:
            pid = int(local_pid.read_text(encoding="utf-8").strip())
        except ValueError:
            pid = -1
        if pid > 0 and (_is_pid_zombie(pid) or not _is_pid_alive(pid)):
            updated = status.model_copy(
                update={
                    "state": "failed",
                    "ended_at": status.ended_at or datetime.now(),
                    "message": "process ended without exit code file",
                }
            )
            if updated != status:
                write_status(status_path, updated)
                resolved = resolve_run_paths(run_dir)
                append_event(resolved, run_id=run_dir.name, event="state_changed", state=updated.state, message=updated.message)
                write_summary(resolved, run_dir.name)
            return updated

    return status


def _collect_rows(project_root: Path, state_filter: str | None, tags: list[str] | None) -> list[dict[str, str]]:
    runs_root = project_root / DEFAULT_RUNS_DIR
    if not runs_root.exists():
        return []

    run_dirs = sorted([p for p in runs_root.iterdir() if p.is_dir()], reverse=True)
    if not run_dirs:
        return []

    selected_tags = sorted({tag.strip() for tag in (tags or []) if tag and tag.strip()})
    active_states = {"pending", "running"}
    rows_with_sort: list[tuple[float, dict[str, str]]] = []

    for run_dir in run_dirs:
        metadata_path = run_dir / "metadata.json"
        status_path = run_dir / "status.json"
        if not metadata_path.exists() or not status_path.exists():
            continue

        metadata = read_metadata(metadata_path)
        status = read_status(status_path)
        if metadata.get("backend_type") == "local":
            status = _refresh_local_status_if_possible(run_dir, status)
        if status.state not in active_states:
            continue
        if state_filter and status.state != state_filter:
            continue

        run_tags = _normalize_tags(metadata.get("tags"))
        if not _matches_tags(run_tags, selected_tags):
            continue

        backend = metadata.get("backend_type", "?")
        if backend == "local":
            pid_file = run_dir / "local.pid"
            if pid_file.exists():
                try:
                    pid = int(pid_file.read_text(encoding="utf-8").strip())
                    pid_or_job = str(pid)
                    alive = "yes" if _is_pid_alive(pid) else "no"
                except ValueError:
                    pid_or_job = "-"
                    alive = "unknown"
            else:
                pid_or_job = "-"
                alive = "unknown"
        else:
            pid_or_job = str(metadata.get("slurm_job_id") or "-")
            alive = "unknown"

        started_sort = status.started_at.timestamp() if status.started_at else float("-inf")
        rows_with_sort.append(
            (
                started_sort,
                {
                    "run_id": run_dir.name,
                    "state": status.state,
                    "backend": backend,
                    "pid": pid_or_job,
                    "alive": alive,
                    "started_at": iso_short(status.started_at),
                    "experiment": str(metadata.get("experiment_name", "?")),
                    "tags": ",".join(run_tags) or "-",
                    "experiment_color": str(metadata.get("experiment_color") or ""),
                },
            )
        )
    rows_with_sort.sort(key=lambda item: item[0], reverse=True)
    return [row for _, row in rows_with_sort]


def _build_lines(rows: list[dict[str, str]], *, use_color: bool) -> list[str]:
    if not rows:
        return ["no active runs matched filter"]
    columns = ["run_id", "state", "backend", "pid", "alive", "started_at", "experiment", "tags"]
    lines = render_table(columns, rows)
    if not use_color:
        return lines
    colored = [lines[0]]
    for idx, line in enumerate(lines[1:]):
        colored.append(colorize(line, rows[idx].get("experiment_color"), enabled=True))
    return colored


def _print_lines(lines: list[str]) -> None:
    for line in lines:
        print(line)


def _redraw_lines_tty(lines: list[str], previous_count: int) -> int:
    if previous_count > 0:
        for _ in range(previous_count):
            sys.stdout.write("\x1b[1A")
            sys.stdout.write("\x1b[2K")
    for line in lines:
        sys.stdout.write(line + "\n")
    sys.stdout.flush()
    return len(lines)


def list_processes(
    *,
    state_filter: str | None = None,
    tags: list[str] | None = None,
    json_output: bool = False,
    watch: bool = False,
    interval_sec: float = 2.0,
) -> int:
    project_root = Path.cwd().resolve()
    use_color = should_use_color()

    rows = _collect_rows(project_root, state_filter, tags)
    if not rows and json_output:
        print(json.dumps(ps_result_payload(items=[])))
        return 0

    if watch and json_output:
        watch = False

    if json_output:
        print(json.dumps(ps_result_payload(items=rows)))
        return 0

    if not watch:
        _print_lines(_build_lines(rows, use_color=use_color))
        return 0

    if not sys.stdout.isatty():
        _print_lines(_build_lines(rows, use_color=use_color))
        return 0

    prev_count = 0
    sys.stdout.write("\x1b[?25l")
    sys.stdout.flush()
    try:
        while True:
            current_rows = _collect_rows(project_root, state_filter, tags)
            lines = _build_lines(current_rows, use_color=use_color)
            prev_count = _redraw_lines_tty(lines, prev_count)
            time.sleep(max(interval_sec, 0.1))
    except KeyboardInterrupt:
        pass
    finally:
        sys.stdout.write("\x1b[?25h")
        sys.stdout.flush()
        return 0



from __future__ import annotations

import shutil
import subprocess
from pathlib import Path


def command_exists(name: str) -> bool:
    return shutil.which(name) is not None


def run_capture(args: list[str], cwd: str | Path | None = None) -> subprocess.CompletedProcess[str]:
    return subprocess.run(
        args,
        cwd=str(cwd) if cwd else None,
        check=False,
        capture_output=True,
        text=True,
    )



from __future__ import annotations

import json
from pathlib import Path

from one_run.backends.base import RunContext
from one_run.backends.slurm import SlurmBackend
from one_run.commands.common import alloc_run_dir
from one_run.commands.deps_cmd import wait_for_dependencies
from one_run.errors import BackendError
from one_run.manifest import load_manifest
from one_run.metadata import build_metadata
from one_run.models import RunRuntime, RunStatus
from one_run.payloads import submit_result_payload
from one_run.runtime import local_runtime_snapshot
from one_run.status_store import write_metadata, write_runtime, write_status
from one_run.tracking import append_event, write_summary


def submit_slurm(
    config_path: str,
    *,
    json_output: bool = False,
    wait_for: list[str] | None = None,
    wait_timeout: str | None = None,
    wait_interval: float | None = None,
) -> int:
    manifest_path = Path(config_path).resolve()
    manifest = load_manifest(manifest_path)

    if manifest.backend.kind != "slurm":
        raise BackendError("this run requires backend.kind=slurm")

    project_root = Path.cwd().resolve()
    wait_for_dependencies(
        manifest,
        project_root=project_root,
        wait_for_run_ids=wait_for,
        wait_timeout=wait_timeout,
        wait_interval=wait_interval,
        emit_console=not json_output,
    )
    run_id, paths = alloc_run_dir(project_root, manifest.experiment.name)
    if not json_output:
        print(f"run_id={run_id}", flush=True)
    paths.config_snapshot.write_text(manifest_path.read_text(encoding="utf-8"), encoding="utf-8")

    metadata = build_metadata(run_id, manifest, project_root, Path.cwd())
    write_metadata(paths.metadata, metadata.model_dump(mode="json"))
    write_runtime(paths.runtime, local_runtime_snapshot(duration_sec=None, peak_rss_raw=None))
    write_status(paths.status, RunStatus(state="pending"))
    append_event(paths, run_id=run_id, event="created", state="pending")
    write_summary(paths, run_id)

    backend = SlurmBackend()
    run_ctx = RunContext(
        run_id=run_id,
        manifest_path=manifest_path,
        manifest=manifest,
        paths=paths,
        project_root=project_root,
        metadata=metadata,
    )

    backend.prepare(run_ctx)
    result = backend.start(run_ctx)

    if result.slurm_job_id:
        metadata = metadata.model_copy(update={"slurm_job_id": result.slurm_job_id})
        write_metadata(paths.metadata, metadata.model_dump(mode="json"))

    write_status(
        paths.status,
        RunStatus(
            state=result.state,
            started_at=result.started_at,
            ended_at=result.ended_at,
            exit_code=result.exit_code,
            message=result.message,
        ),
    )
    if result.runtime:
        write_runtime(paths.runtime, RunRuntime.model_validate(result.runtime))
    append_event(paths, run_id=run_id, event="submitted", state=result.state, message=result.message)
    write_summary(paths, run_id)

    if json_output:
        print(json.dumps(submit_result_payload(
            run_id=run_id,
            state=result.state,
            slurm_job_id=result.slurm_job_id,
            exit_code=result.exit_code,
            message=result.message,
            run_dir=str(paths.run_dir),
            summary=str(paths.summary),
            events=str(paths.events),
        )))
    else:
        if result.slurm_job_id:
            print(f"slurm_job_id={result.slurm_job_id}")
        print(f"state={result.state}")
    return 0 if result.state in {"pending", "running", "success"} else 1



from __future__ import annotations

import subprocess
import resource
import sys
import threading
import shlex
from pathlib import Path
from datetime import datetime
from typing import BinaryIO

from one_run.backends.base import BackendQueryResult, BackendStartResult, RunContext
from one_run.errors import BackendError
from one_run.runtime import local_runtime_snapshot
from one_run.utils.shell import command_exists, run_capture
from one_run.utils.templating import render_command


class LocalBackend:
    @staticmethod
    def _tee_stream(pipe: BinaryIO, log_file: BinaryIO, console: BinaryIO | None) -> None:
        while True:
            chunk = pipe.readline()
            if not chunk:
                break
            log_file.write(chunk)
            log_file.flush()
            if console is not None:
                console.write(chunk)
                console.flush()

    def _render(self, run_ctx: RunContext) -> tuple[str, str | None]:
        if run_ctx.manifest.environment.kind == "docker":
            return self._render_docker(run_ctx)

        return (
            render_command(
                run_ctx.manifest.entrypoint.command,
                {
                    "config": str(run_ctx.paths.config_snapshot),
                    "run_dir": str(run_ctx.paths.run_dir),
                    "seed": run_ctx.manifest.experiment.seed,
                    "project_root": str(run_ctx.project_root),
                },
            ),
            None,
        )

    @staticmethod
    def _docker_image_exists(image: str) -> bool:
        proc = run_capture(["docker", "image", "inspect", image])
        return proc.returncode == 0

    @staticmethod
    def _docker_pull(image: str) -> None:
        proc = run_capture(["docker", "pull", image])
        if proc.returncode != 0:
            msg = (proc.stderr or proc.stdout).strip() or "docker pull failed"
            raise BackendError(f"docker pull failed for image '{image}': {msg}")

    @staticmethod
    def _docker_image_digest(image: str) -> str | None:
        proc = run_capture(["docker", "image", "inspect", "--format", "{{index .RepoDigests 0}}", image])
        if proc.returncode != 0:
            return None
        digest = proc.stdout.strip()
        return digest or None

    @staticmethod
    def _resolve_env_file(path_str: str, project_root: Path) -> str:
        p = Path(path_str)
        if not p.is_absolute():
            p = project_root / p
        return str(p)

    def _ensure_docker_image(self, run_ctx: RunContext, image: str) -> str | None:
        env_cfg = run_ctx.manifest.environment
        pull_policy = env_cfg.pull_policy or "if_missing"
        exists = self._docker_image_exists(image)

        if pull_policy == "always":
            self._docker_pull(image)
        elif pull_policy == "if_missing":
            if not exists:
                self._docker_pull(image)
        elif pull_policy == "never":
            if not exists:
                raise BackendError(
                    f"docker image '{image}' not found locally and pull_policy=never"
                )

        return self._docker_image_digest(image)

    def _render_docker(self, run_ctx: RunContext) -> tuple[str, str | None]:
        env_cfg = run_ctx.manifest.environment
        image = env_cfg.image
        if not image:
            raise BackendError("environment.image is required when environment.kind=docker")

        digest = self._ensure_docker_image(run_ctx, image)

        container_project_root = "/workspace"
        container_run_dir = "/run"
        container_config = f"{container_run_dir}/config.snapshot.yaml"
        workdir = env_cfg.workdir or container_project_root

        inner_command = render_command(
            run_ctx.manifest.entrypoint.command,
            {
                "config": container_config,
                "run_dir": container_run_dir,
                "seed": run_ctx.manifest.experiment.seed,
                "project_root": container_project_root,
            },
        )

        cmd_parts = [
            "docker",
            "run",
            "--rm",
            "-v",
            f"{run_ctx.project_root}:{container_project_root}",
            "-v",
            f"{run_ctx.paths.run_dir}:{container_run_dir}",
            "-w",
            workdir,
        ]

        if env_cfg.mounts:
            for mount in env_cfg.mounts:
                cmd_parts.extend(["-v", mount])
        if env_cfg.env_file:
            cmd_parts.extend(["--env-file", self._resolve_env_file(env_cfg.env_file, run_ctx.project_root)])
        if env_cfg.env:
            for key, value in env_cfg.env.items():
                cmd_parts.extend(["-e", f"{key}={value}"])

        labels = {
            "one_run_id": run_ctx.run_id,
            "one_run_experiment": run_ctx.manifest.experiment.name,
            "one_run_backend": run_ctx.manifest.backend.kind,
        }
        if env_cfg.labels:
            labels.update(env_cfg.labels)
        for key, value in labels.items():
            cmd_parts.extend(["--label", f"{key}={value}"])

        resources = run_ctx.manifest.resources
        if resources and resources.gpus:
            cmd_parts.extend(["--gpus", str(resources.gpus)])

        cmd_parts.extend([image, "/bin/sh", "-lc", inner_command])
        return " ".join(shlex.quote(part) for part in cmd_parts), digest

    def prepare(self, run_ctx: RunContext) -> None:
        if run_ctx.manifest.environment.kind == "docker" and not command_exists("docker"):
            raise BackendError("docker not found. local docker execution is unavailable")
        return

    def start(self, run_ctx: RunContext, *, emit_console: bool = True) -> BackendStartResult:
        command, image_digest = self._render(run_ctx)

        started_at = datetime.now()
        with run_ctx.paths.stdout_log.open("ab") as stdout_f, run_ctx.paths.stderr_log.open("ab") as stderr_f:
            process = subprocess.Popen(
                command,
                shell=True,
                cwd=str(run_ctx.project_root),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            run_ctx.paths.local_pid.write_text(f"{process.pid}\n", encoding="utf-8")
            stdout_pipe = process.stdout
            stderr_pipe = process.stderr
            if stdout_pipe is None or stderr_pipe is None:
                raise RuntimeError("failed to capture process streams")

            out_thread = threading.Thread(
                target=self._tee_stream,
                args=(stdout_pipe, stdout_f, sys.stdout.buffer if emit_console else None),
                daemon=True,
            )
            err_thread = threading.Thread(
                target=self._tee_stream,
                args=(stderr_pipe, stderr_f, sys.stderr.buffer if emit_console else None),
                daemon=True,
            )
            out_thread.start()
            err_thread.start()
            try:
                exit_code = process.wait()
            except KeyboardInterrupt:
                process.terminate()
                try:
                    process.wait(timeout=10)
                except subprocess.TimeoutExpired:
                    process.kill()
                    process.wait()
                out_thread.join(timeout=2)
                err_thread.join(timeout=2)
                run_ctx.paths.local_exit_code.write_text("130\n", encoding="utf-8")
                return BackendStartResult(
                    state="failed",
                    started_at=started_at,
                    ended_at=datetime.now(),
                    exit_code=130,
                    message="interrupted",
                    docker_image_digest=image_digest,
                    runtime=local_runtime_snapshot(
                        duration_sec=(datetime.now() - started_at).total_seconds(),
                        peak_rss_raw=resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss,
                    ).model_dump(mode="json"),
                )
            out_thread.join()
            err_thread.join()
            run_ctx.paths.local_exit_code.write_text(f"{exit_code}\n", encoding="utf-8")

        ended_at = datetime.now()
        rss_after = resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
        state = "success" if exit_code == 0 else "failed"
        return BackendStartResult(
            state=state,
            started_at=started_at,
            ended_at=ended_at,
            exit_code=exit_code,
            message=None if state == "success" else f"command exited with code {exit_code}",
            docker_image_digest=image_digest,
            runtime=local_runtime_snapshot(
                duration_sec=(ended_at - started_at).total_seconds(),
                peak_rss_raw=rss_after,
            ).model_dump(mode="json"),
        )

    def start_detached(self, run_ctx: RunContext) -> BackendStartResult:
        command, image_digest = self._render(run_ctx)
        started_at = datetime.now()
        wrapper = (
            f"{command}; rc=$?; "
            f"echo $rc > '{run_ctx.paths.local_exit_code}'"
        )
        with run_ctx.paths.stdout_log.open("ab") as stdout_f, run_ctx.paths.stderr_log.open("ab") as stderr_f:
            process = subprocess.Popen(
                ["/bin/sh", "-c", wrapper],
                cwd=str(run_ctx.project_root),
                stdout=stdout_f,
                stderr=stderr_f,
                start_new_session=True,
            )
        run_ctx.paths.local_pid.write_text(f"{process.pid}\n", encoding="utf-8")
        return BackendStartResult(
            state="running",
            started_at=started_at,
            ended_at=None,
            exit_code=None,
            message=None,
            docker_image_digest=image_digest,
        )

    def query(self, run_ctx: RunContext, status):
        return BackendQueryResult(state=status.state, exit_code=status.exit_code, message=status.message)

    def cancel(self, run_ctx: RunContext) -> None:
        return



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from one_run.run_id import make_run_id
from one_run.constants import (
    ARTIFACTS_DIR_NAME,
    CONFIG_SNAPSHOT_NAME,
    LOGS_DIR_NAME,
    METADATA_NAME,
    METRICS_DIR_NAME,
    LOCAL_EXIT_CODE_NAME,
    LOCAL_PID_NAME,
    RUNTIME_NAME,
    SUMMARY_NAME,
    SUMMARY_MIN_NAME,
    EVENTS_NAME,
    STDERR_LOG_NAME,
    STATUS_NAME,
    STDOUT_LOG_NAME,
)


@dataclass(frozen=True)
class RunPaths:
    run_dir: Path
    config_snapshot: Path
    metadata: Path
    status: Path
    runtime: Path
    summary: Path
    summary_min: Path
    events: Path
    local_pid: Path
    local_exit_code: Path
    logs_dir: Path
    stdout_log: Path
    stderr_log: Path
    artifacts_dir: Path
    metrics_dir: Path


def resolve_run_paths(run_dir: Path) -> RunPaths:
    logs = run_dir / LOGS_DIR_NAME
    artifacts = run_dir / ARTIFACTS_DIR_NAME
    metrics = run_dir / METRICS_DIR_NAME

    return RunPaths(
        run_dir=run_dir,
        config_snapshot=run_dir / CONFIG_SNAPSHOT_NAME,
        metadata=run_dir / METADATA_NAME,
        status=run_dir / STATUS_NAME,
        runtime=run_dir / RUNTIME_NAME,
        summary=run_dir / SUMMARY_NAME,
        summary_min=run_dir / SUMMARY_MIN_NAME,
        events=run_dir / EVENTS_NAME,
        local_pid=run_dir / LOCAL_PID_NAME,
        local_exit_code=run_dir / LOCAL_EXIT_CODE_NAME,
        logs_dir=logs,
        stdout_log=logs / STDOUT_LOG_NAME,
        stderr_log=logs / STDERR_LOG_NAME,
        artifacts_dir=artifacts,
        metrics_dir=metrics,
    )


def create_run_paths(run_dir: Path) -> RunPaths:
    paths = resolve_run_paths(run_dir)

    run_dir.mkdir(parents=True, exist_ok=False)
    paths.logs_dir.mkdir(parents=True, exist_ok=True)
    paths.artifacts_dir.mkdir(parents=True, exist_ok=True)
    paths.metrics_dir.mkdir(parents=True, exist_ok=True)
    return paths


def allocate_run_paths(runs_root: Path, experiment_name: str, *, max_attempts: int = 100) -> tuple[str, RunPaths]:
    base_run_id = make_run_id(experiment_name)
    for attempt in range(max_attempts):
        run_id = base_run_id if attempt == 0 else f"{base_run_id}-{attempt:02d}"
        try:
            return run_id, create_run_paths(runs_root / run_id)
        except FileExistsError:
            continue
    raise FileExistsError(f"failed to allocate unique run_id after {max_attempts} attempts")



from __future__ import annotations

import os
import time
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from one_run.constants import DEFAULT_RUNS_DIR
from one_run.errors import OneRunError
from one_run.models import Manifest
from one_run.status_store import read_metadata, read_status
from one_run.utils.duration import parse_duration

_TERMINAL_STATES = {"success", "failed"}


@dataclass
class _WaitConfig:
    run_ids: list[str]
    tags: list[str]
    timeout_sec: float | None
    interval_sec: float


def _state_for_run(run_dir: Path) -> str:
    status = read_status(run_dir / "status.json")
    if status.state in _TERMINAL_STATES:
        return status.state

    local_exit = run_dir / "local.exit_code"
    if local_exit.exists():
        try:
            code = int(local_exit.read_text(encoding="utf-8").strip())
        except ValueError:
            code = 1
        return "success" if code == 0 else "failed"

    local_pid = run_dir / "local.pid"
    if local_pid.exists():
        try:
            pid = int(local_pid.read_text(encoding="utf-8").strip())
            os.kill(pid, 0)
        except (ValueError, OSError):
            return "failed"

    return status.state


def _resolve_wait_config(
    manifest: Manifest,
    *,
    wait_for_run_ids: list[str] | None,
    wait_timeout: str | None,
    wait_interval: float | None,
) -> _WaitConfig:
    dep = manifest.depends_on
    run_ids = list(wait_for_run_ids or [])
    tags: list[str] = []
    timeout_str = wait_timeout
    interval_sec = wait_interval

    if dep:
        run_ids.extend(dep.run_ids or [])
        tags.extend(dep.tags or [])
        if timeout_str is None:
            timeout_str = dep.timeout
        if interval_sec is None:
            interval_sec = dep.interval_sec

    run_ids = sorted({value.strip() for value in run_ids if value and value.strip()})
    tags = sorted({value.strip() for value in tags if value and value.strip()})

    timeout_sec: float | None = None
    if timeout_str:
        try:
            timeout_sec = parse_duration(timeout_str).total_seconds()
        except ValueError as exc:
            raise OneRunError(str(exc)) from exc

    resolved_interval = interval_sec if interval_sec is not None else 2.0
    if resolved_interval <= 0:
        raise OneRunError("wait interval must be > 0")

    return _WaitConfig(
        run_ids=run_ids,
        tags=tags,
        timeout_sec=timeout_sec,
        interval_sec=resolved_interval,
    )


def _tag_status(project_root: Path, tag: str) -> str:
    runs_root = project_root / DEFAULT_RUNS_DIR
    if not runs_root.exists():
        return "missing"

    found = False
    has_running = False
    has_failed = False
    for run_dir in runs_root.iterdir():
        if not run_dir.is_dir():
            continue
        metadata_path = run_dir / "metadata.json"
        status_path = run_dir / "status.json"
        if not metadata_path.exists() or not status_path.exists():
            continue
        metadata = read_metadata(metadata_path)
        raw_tags = metadata.get("tags")
        if not isinstance(raw_tags, list):
            continue
        tags = {str(value).strip() for value in raw_tags if str(value).strip()}
        if tag not in tags:
            continue
        found = True
        state = _state_for_run(run_dir)
        if state == "success":
            return "ready"
        if state not in _TERMINAL_STATES:
            has_running = True
        elif state == "failed":
            has_failed = True

    if not found:
        return "missing"
    if has_running:
        return "running"
    if has_failed:
        return "failed"
    return "running"


def wait_for_dependencies(
    manifest: Manifest,
    *,
    project_root: Path,
    wait_for_run_ids: list[str] | None = None,
    wait_timeout: str | None = None,
    wait_interval: float | None = None,
    emit_console: bool = True,
) -> None:
    cfg = _resolve_wait_config(
        manifest,
        wait_for_run_ids=wait_for_run_ids,
        wait_timeout=wait_timeout,
        wait_interval=wait_interval,
    )
    if not cfg.run_ids and not cfg.tags:
        return

    started_at = datetime.now()
    if emit_console:
        print("waiting for dependencies...", flush=True)

    while True:
        all_ready = True

        for run_id in cfg.run_ids:
            run_dir = project_root / DEFAULT_RUNS_DIR / run_id
            if not run_dir.exists():
                all_ready = False
                continue
            state = _state_for_run(run_dir)
            if state == "failed":
                raise OneRunError(f"dependency failed: run_id={run_id}")
            if state != "success":
                all_ready = False

        for tag in cfg.tags:
            tag_status = _tag_status(project_root, tag)
            if tag_status == "ready":
                continue
            if tag_status == "failed":
                raise OneRunError(f"dependency failed: tag={tag}")
            if tag_status in {"running", "missing"}:
                all_ready = False
                continue
            all_ready = False

        if all_ready:
            if emit_console:
                print("dependencies ready", flush=True)
            return

        if cfg.timeout_sec is not None:
            elapsed = (datetime.now() - started_at).total_seconds()
            if elapsed >= cfg.timeout_sec:
                raise OneRunError("dependency wait timed out")

        time.sleep(cfg.interval_sec)



from one_run.backends.base import Backend, BackendQueryResult, BackendStartResult, RunContext
from one_run.backends.local import LocalBackend
from one_run.backends.slurm import SlurmBackend

__all__ = [
    "Backend",
    "BackendQueryResult",
    "BackendStartResult",
    "RunContext",
    "LocalBackend",
    "SlurmBackend",
]



from __future__ import annotations

from pathlib import Path

import yaml
from pydantic import ValidationError

from one_run.compat import warn_legacy_manifest
from one_run.errors import ManifestError
from one_run.models import Manifest


def load_manifest(path: str | Path) -> Manifest:
    manifest_path = Path(path)
    if not manifest_path.exists():
        raise ManifestError(f"manifest not found: {manifest_path}")

    try:
        payload = yaml.safe_load(manifest_path.read_text(encoding="utf-8"))
    except yaml.YAMLError as exc:
        raise ManifestError(f"invalid yaml in {manifest_path}: {exc}") from exc

    if payload is None:
        raise ManifestError(f"manifest is empty: {manifest_path}")

    warn_legacy_manifest(payload, manifest_path)

    try:
        return Manifest.model_validate(payload)
    except ValidationError as exc:
        raise ManifestError(f"manifest validation failed for {manifest_path}:\n{exc}") from exc



from __future__ import annotations

from datetime import datetime
from typing import Any


def iso_short(value: Any) -> str:
    if value is None:
        return "-"
    if isinstance(value, datetime):
        return value.replace(microsecond=0).isoformat()
    if isinstance(value, str):
        parsed = datetime.fromisoformat(value)
        return parsed.replace(microsecond=0).isoformat()
    return str(value)



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Callable


LineHandler = Callable[[Path, str, dict[str, Any] | None], None]


def consume_event_file(path: Path, *, offset: int, on_line: LineHandler) -> int:
    with path.open("r", encoding="utf-8", errors="replace") as f:
        f.seek(offset)
        for line in f:
            text = line.rstrip("\n")
            if not text:
                continue
            payload: dict[str, Any] | None
            try:
                raw = json.loads(text)
            except json.JSONDecodeError:
                payload = None
            else:
                payload = raw if isinstance(raw, dict) else None
            on_line(path, text, payload)
        return f.tell()


def ensure_offset(path: Path, *, offsets: dict[Path, int], history: bool) -> int:
    if path not in offsets:
        offsets[path] = 0 if history else path.stat().st_size
    return offsets[path]



from __future__ import annotations

import string

from one_run.constants import ALLOWED_PLACEHOLDERS
from one_run.errors import TemplateError


def validate_placeholders(template: str) -> None:
    formatter = string.Formatter()
    for _, field_name, _, _ in formatter.parse(template):
        if field_name is None:
            continue
        if field_name not in ALLOWED_PLACEHOLDERS:
            allowed = ", ".join(sorted(ALLOWED_PLACEHOLDERS))
            raise TemplateError(f"unknown placeholder '{field_name}'. allowed: {allowed}")


def render_command(template: str, values: dict[str, str | int]) -> str:
    validate_placeholders(template)
    try:
        return template.format(**values)
    except KeyError as exc:
        raise TemplateError(f"missing value for placeholder: {exc.args[0]}") from exc



from __future__ import annotations

import re
from pathlib import Path

import yaml

from one_run.errors import OneRunError


def _ask(prompt: str, *, default: str | None = None) -> str:
    suffix = f" [{default}]" if default is not None else ""
    raw = input(f"{prompt}{suffix}: ").strip()
    if raw:
        return raw
    if default is not None:
        return default
    return ""


def _ask_choice(prompt: str, *, choices: tuple[str, ...], default: str) -> str:
    options = "/".join(choices)
    while True:
        value = _ask(f"{prompt} ({options})", default=default).lower()
        if value in choices:
            return value
        print(f"invalid choice: {value}. expected one of: {options}")


def _ask_int(prompt: str, *, default: int | None = None, required: bool = False) -> int | None:
    default_text = str(default) if default is not None else None
    while True:
        raw = _ask(prompt, default=default_text)
        if not raw:
            if required:
                print("value is required")
                continue
            return None
        try:
            return int(raw)
        except ValueError:
            print("expected integer value")


def _ask_optional_csv(prompt: str) -> list[str] | None:
    raw = _ask(prompt, default="")
    if not raw:
        return None
    values = [part.strip() for part in raw.split(",")]
    tags = [value for value in values if value]
    return tags or None


def _slugify_filename(value: str) -> str:
    slug = re.sub(r"[^a-zA-Z0-9._-]+", "-", value.strip().lower()).strip("-")
    return slug or "manifest"


def init_manifest(*, output: str | None, force: bool = False) -> int:

    print("one-run init manifest")
    print("press enter to accept default values")

    experiment_name = _ask("experiment name", default="local-test")
    seed = _ask_int("seed", default=42, required=True)
    tags = _ask_optional_csv("tags (comma-separated, optional)")
    color = _ask("experiment color (optional)", default="").lower() or None
    command = _ask("entrypoint command", default='python3 -c "print(\\"ok\\")"')

    environment_kind = _ask_choice("environment kind", choices=("local", "docker"), default="local")
    backend_kind = _ask_choice("backend kind", choices=("local", "slurm"), default="local")

    cpus = _ask_int("resources.cpus (optional)")
    mem_gb = _ask_int("resources.mem_gb (optional)")
    gpus = _ask_int("resources.gpus (optional)")
    time_min = _ask_int("resources.time_min (optional)")

    manifest: dict = {
        "experiment": {
            "name": experiment_name,
            "seed": seed,
        },
        "entrypoint": {
            "command": command,
        },
        "environment": {
            "kind": environment_kind,
        },
        "backend": {
            "kind": backend_kind,
        },
    }

    if tags:
        manifest["experiment"]["tags"] = tags
    if color:
        manifest["experiment"]["color"] = color

    if environment_kind == "docker":
        image = _ask("environment.image", default="python:3.11-slim")
        workdir = _ask("environment.workdir (optional)", default="") or None
        mounts = _ask_optional_csv("environment.mounts (comma-separated host:container, optional)")
        manifest["environment"]["image"] = image
        if workdir:
            manifest["environment"]["workdir"] = workdir
        if mounts:
            manifest["environment"]["mounts"] = mounts

    resources = {k: v for k, v in {
        "cpus": cpus,
        "mem_gb": mem_gb,
        "gpus": gpus,
        "time_min": time_min,
    }.items() if v is not None}
    if resources:
        manifest["resources"] = resources

    if backend_kind == "slurm":
        partition = _ask("backend.slurm.partition (optional)", default="") or None
        manifest["backend"]["slurm"] = {}
        if partition:
            manifest["backend"]["slurm"]["partition"] = partition

    if output:
        output_path = Path(output).resolve()
    else:
        filename = f"{_slugify_filename(experiment_name)}.manifest.yaml"
        output_path = (Path.cwd() / "manifests" / filename).resolve()

    if output_path.exists() and not force:
        raise OneRunError(f"file already exists: {output_path} (use --force to overwrite)")

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(yaml.safe_dump(manifest, sort_keys=False), encoding="utf-8")

    print(f"written={output_path}")
    return 0


def init_listener(*, output: str | None, force: bool = False) -> int:
    print("one-run init listener")
    print("press enter to accept default values")

    kind = _ask_choice("listener kind", choices=("local", "webhook"), default="local")
    name = _ask("listener name", default="listener-profile")
    interval_sec = _ask_int("interval_sec (optional)", default=1, required=False)
    events = _ask_optional_csv("events (comma-separated, optional)")
    states = _ask_optional_csv("states (comma-separated, optional)")
    tags = _ask_optional_csv("tags (comma-separated, optional)")

    listener: dict = {
        "kind": kind,
        "name": name,
    }

    if events:
        listener["events"] = events
    if states:
        listener["states"] = states
    if tags:
        listener["tags"] = tags

    if kind == "local":
        handler = _ask("handler module:function (optional)", default="") or None
        if handler:
            listener["handler"] = handler
    else:
        webhook_url = _ask("webhook_url", default="http://127.0.0.1:8088/events")
        action = _ask("action (optional)", default="") or None
        timeout_sec = _ask_int("timeout_sec (optional)", default=3, required=False)
        listener["webhook_url"] = webhook_url
        if action:
            listener["action"] = action
        if timeout_sec is not None:
            listener["timeout_sec"] = timeout_sec

    payload: dict = {
        "listeners": [listener],
    }
    if interval_sec is not None:
        payload["interval_sec"] = interval_sec

    if output:
        output_path = Path(output).resolve()
    else:
        filename = f"{_slugify_filename(name)}.listener.yaml"
        output_path = (Path.cwd() / "listeners" / filename).resolve()

    if output_path.exists() and not force:
        raise OneRunError(f"file already exists: {output_path} (use --force to overwrite)")

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(yaml.safe_dump(payload, sort_keys=False), encoding="utf-8")
    print(f"written={output_path}")
    return 0



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Protocol

from one_run.models import Manifest, RunMetadata, RunStatus
from one_run.run_dir import RunPaths


@dataclass(frozen=True)
class RunContext:
    run_id: str
    manifest_path: Path
    manifest: Manifest
    paths: RunPaths
    project_root: Path
    metadata: RunMetadata


@dataclass(frozen=True)
class BackendStartResult:
    state: str
    started_at: datetime | None = None
    ended_at: datetime | None = None
    exit_code: int | None = None
    message: str | None = None
    slurm_job_id: str | None = None
    docker_image_digest: str | None = None
    runtime: dict[str, Any] | None = None


@dataclass(frozen=True)
class BackendQueryResult:
    state: str
    exit_code: int | None = None
    message: str | None = None
    runtime: dict[str, Any] | None = None


class Backend(Protocol):
    def prepare(self, run_ctx: RunContext) -> None:
        ...

    def start(self, run_ctx: RunContext) -> BackendStartResult:
        ...

    def query(self, run_ctx: RunContext, status: RunStatus) -> BackendQueryResult | None:
        ...

    def cancel(self, run_ctx: RunContext) -> None:
        ...



from __future__ import annotations

from typing import Any

from one_run.constants import JSON_SCHEMA_VERSION
from one_run.models import RunMetadata, RunRuntime, RunStatus


def run_result_payload(
    *,
    run_id: str,
    state: str,
    watch: bool,
    exit_code: int | None,
    message: str | None,
    run_dir: str,
    summary: str,
    events: str,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "run_result",
        "run_id": run_id,
        "state": state,
        "watch": watch,
        "exit_code": exit_code,
        "message": message,
        "paths": {
            "run_dir": run_dir,
            "summary": summary,
            "events": events,
        },
    }


def submit_result_payload(
    *,
    run_id: str,
    state: str,
    slurm_job_id: str | None,
    exit_code: int | None,
    message: str | None,
    run_dir: str,
    summary: str,
    events: str,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "submit_result",
        "run_id": run_id,
        "state": state,
        "slurm_job_id": slurm_job_id,
        "exit_code": exit_code,
        "message": message,
        "paths": {
            "run_dir": run_dir,
            "summary": summary,
            "events": events,
        },
    }


def status_result_payload(
    *,
    run_id: str,
    wait: bool,
    timed_out: bool,
    backend: str,
    metadata: RunMetadata,
    status: RunStatus,
    runtime: RunRuntime | None,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "status_result",
        "run_id": run_id,
        "wait": wait,
        "timed_out": timed_out,
        "backend": backend,
        "metadata": metadata.model_dump(mode="json"),
        "status": status.model_dump(mode="json"),
        "runtime": runtime.model_dump(mode="json") if runtime else None,
    }


def validate_result_payload(
    *,
    manifest: str,
    valid: bool,
    backend: str,
    environment: str,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "validate_result",
        "manifest": manifest,
        "valid": valid,
        "backend": backend,
        "environment": environment,
    }


def cancel_result_payload(
    *,
    run_id: str,
    state: str,
    force: bool,
    cancelled: bool,
    exit_code: int | None = None,
    message: str | None = None,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "cancel_result",
        "run_id": run_id,
        "state": state,
        "force": force,
        "cancelled": cancelled,
        "exit_code": exit_code,
        "message": message,
    }


def rm_cleanup_result_payload(*, items: list[dict[str, str]], count: int, dry_run: bool) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "rm_cleanup_result",
        "items": items,
        "count": count,
        "dry_run": dry_run,
    }


def ls_result_payload(*, items: list[dict[str, Any]]) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "ls_result",
        "items": items,
    }


def ps_result_payload(*, items: list[dict[str, str]]) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "ps_result",
        "items": items,
    }


def rm_result_payload(
    *,
    run_id: str,
    removed: bool,
    cancelled: bool,
    force: bool,
    state_before: str,
    state_after: str | None = None,
) -> dict[str, Any]:
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "kind": "rm_result",
        "run_id": run_id,
        "removed": removed,
        "cancelled": cancelled,
        "force": force,
        "state_before": state_before,
        "state_after": state_after,
    }



from __future__ import annotations

import time
from collections import deque
from datetime import datetime
from pathlib import Path

from one_run.utils.duration import parse_duration
from one_run.constants import DEFAULT_RUNS_DIR
from one_run.errors import OneRunError


def _log_path(run_id: str, project_root: Path, stderr: bool) -> Path:
    run_dir = project_root / DEFAULT_RUNS_DIR / run_id
    if not run_dir.exists():
        raise OneRunError(f"run not found: {run_id}")
    path = run_dir / "logs" / ("stderr.log" if stderr else "stdout.log")
    if not path.exists():
        raise OneRunError(f"log file not found: {path}")
    return path


def _print_last_lines(path: Path, n: int) -> None:
    with path.open("r", encoding="utf-8", errors="replace") as f:
        lines = deque(f, maxlen=n)
    for line in lines:
        print(line, end="")


def _follow(path: Path, poll_sec: float = 0.5) -> None:
    with path.open("r", encoding="utf-8", errors="replace") as f:
        f.seek(0, 2)
        while True:
            line = f.readline()
            if line:
                print(line, end="")
                continue
            time.sleep(poll_sec)


def tail_run_log(
    run_id: str,
    *,
    stderr: bool = False,
    follow: bool = False,
    lines: int = 30,
    since: str | None = None,
) -> int:
    project_root = Path.cwd().resolve()
    path = _log_path(run_id, project_root, stderr)

    if since:
        try:
            cutoff = datetime.now().timestamp() - parse_duration(since).total_seconds()
        except ValueError as exc:
            raise OneRunError(str(exc)) from exc

        if path.stat().st_mtime >= cutoff:
            _print_last_lines(path, lines)
    else:
        _print_last_lines(path, lines)

    if follow:
        try:
            _follow(path)
        except KeyboardInterrupt:
            return 0
    return 0



from __future__ import annotations

import os
import sys
from datetime import datetime

from one_run.models import RunRuntime


def _total_ram_gb() -> float | None:
    if not hasattr(os, "sysconf"):
        return None
    keys = os.sysconf_names
    if "SC_PHYS_PAGES" not in keys or "SC_PAGE_SIZE" not in keys:
        return None
    try:
        pages = os.sysconf("SC_PHYS_PAGES")
        page_size = os.sysconf("SC_PAGE_SIZE")
    except (ValueError, OSError):
        return None
    if pages <= 0 or page_size <= 0:
        return None
    return round((pages * page_size) / (1024**3), 2)


def _ru_maxrss_to_mb(value: int) -> float:
    if sys.platform == "darwin":
        return round(value / (1024**2), 2)
    return round(value / 1024, 2)


def local_runtime_snapshot(duration_sec: float | None, peak_rss_raw: int | None) -> RunRuntime:
    peak_rss_mb = _ru_maxrss_to_mb(peak_rss_raw) if peak_rss_raw is not None else None
    return RunRuntime(
        collected_at=datetime.now(),
        duration_sec=round(duration_sec, 3) if duration_sec is not None else None,
        peak_rss_mb=peak_rss_mb,
        cpu_count=os.cpu_count(),
        total_ram_gb=_total_ram_gb(),
    )


def slurm_runtime_snapshot(
    *,
    elapsed: str | None,
    max_rss: str | None,
    alloc_cpus: int | None,
) -> RunRuntime:
    return RunRuntime(
        collected_at=datetime.now(),
        slurm_elapsed=elapsed,
        slurm_max_rss=max_rss,
        slurm_alloc_cpus=alloc_cpus,
    )



from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

from one_run.backends.base import BackendQueryResult, BackendStartResult, RunContext
from one_run.constants import JOB_SCRIPT_NAME
from one_run.errors import BackendError
from one_run.utils.shell import command_exists, run_capture
from one_run.utils.templating import render_command


SLURM_STATE_MAP = {
    "PENDING": "pending",
    "CONFIGURING": "pending",
    "RUNNING": "running",
    "COMPLETING": "running",
    "COMPLETED": "success",
    "FAILED": "failed",
    "CANCELLED": "failed",
    "TIMEOUT": "failed",
    "NODE_FAIL": "failed",
    "PREEMPTED": "failed",
    "OUT_OF_MEMORY": "failed",
}


class SlurmBackend:
    def prepare(self, run_ctx: RunContext) -> None:
        if not command_exists("sbatch"):
            raise BackendError("sbatch not found. slurm backend is unavailable")

    def _build_sbatch_script(self, run_ctx: RunContext) -> Path:
        resources = run_ctx.manifest.resources
        slurm_cfg = run_ctx.manifest.backend.slurm
        lines = ["#!/usr/bin/env bash", "set -euo pipefail"]

        if slurm_cfg and slurm_cfg.partition:
            lines.append(f"#SBATCH --partition={slurm_cfg.partition}")
        if resources and resources.gpus:
            lines.append(f"#SBATCH --gpus={resources.gpus}")
        if resources and resources.cpus:
            lines.append(f"#SBATCH --cpus-per-task={resources.cpus}")
        if resources and resources.mem_gb:
            lines.append(f"#SBATCH --mem={resources.mem_gb}G")
        if resources and resources.time_min:
            lines.append(f"#SBATCH --time={resources.time_min}")

        lines.append(f"#SBATCH --output={run_ctx.paths.stdout_log}")
        lines.append(f"#SBATCH --error={run_ctx.paths.stderr_log}")
        lines.append("")

        command = render_command(
            run_ctx.manifest.entrypoint.command,
            {
                "config": str(run_ctx.paths.config_snapshot),
                "run_dir": str(run_ctx.paths.run_dir),
                "seed": run_ctx.manifest.experiment.seed,
                "project_root": str(run_ctx.project_root),
            },
        )

        lines.append(f"cd {run_ctx.project_root}")
        lines.append(command)

        script_path = run_ctx.paths.run_dir / JOB_SCRIPT_NAME
        script_path.write_text("\n".join(lines) + "\n", encoding="utf-8")
        return script_path

    def start(self, run_ctx: RunContext) -> BackendStartResult:
        script_path = self._build_sbatch_script(run_ctx)
        proc = run_capture(["sbatch", str(script_path)], cwd=run_ctx.project_root)
        if proc.returncode != 0:
            msg = (proc.stderr or proc.stdout).strip() or "sbatch failed"
            return BackendStartResult(
                state="failed",
                started_at=datetime.now(),
                ended_at=datetime.now(),
                exit_code=proc.returncode,
                message=msg,
            )

        output = proc.stdout.strip()
        job_id = self.parse_sbatch_job_id(output)
        if not job_id:
            raise BackendError(f"could not parse slurm job id from sbatch output: {output}")

        return BackendStartResult(
            state="pending",
            started_at=datetime.now(),
            ended_at=None,
            exit_code=None,
            message=None,
            slurm_job_id=job_id,
        )

    @staticmethod
    def parse_sbatch_job_id(output: str) -> str | None:
        match = re.search(r"Submitted batch job\s+(\d+)", output.strip())
        if not match:
            return None
        return match.group(1)

    def _query_squeue(self, job_id: str) -> str | None:
        if not command_exists("squeue"):
            return None
        proc = run_capture(["squeue", "-h", "-j", job_id, "-o", "%T"])
        if proc.returncode != 0:
            return None
        state = proc.stdout.strip().splitlines()
        if not state:
            return None
        return state[0].strip().upper()

    def _query_sacct_row(self, job_id: str) -> dict[str, str] | None:
        if not command_exists("sacct"):
            return None
        proc = run_capture(
            [
                "sacct",
                "-n",
                "-j",
                job_id,
                "--format=State,Elapsed,MaxRSS,AllocCPUS,ExitCode",
                "--parsable2",
            ]
        )
        if proc.returncode != 0:
            return None
        lines = [ln.strip() for ln in proc.stdout.splitlines() if ln.strip()]
        if not lines:
            return None
        parts = lines[0].split("|")
        row = {
            "state": parts[0].strip().upper() if len(parts) > 0 else "",
            "elapsed": parts[1].strip() if len(parts) > 1 else "",
            "max_rss": parts[2].strip() if len(parts) > 2 else "",
            "alloc_cpus": parts[3].strip() if len(parts) > 3 else "",
            "exit_code": parts[4].strip() if len(parts) > 4 else "",
        }
        if not row["state"]:
            return None
        row["state"] = row["state"].split("+")[0]
        return row

    def query(self, run_ctx: RunContext, status) -> BackendQueryResult | None:
        job_id = run_ctx.metadata.slurm_job_id
        if not job_id:
            return None

        sacct_row = self._query_sacct_row(job_id)
        raw_state = self._query_squeue(job_id) or (sacct_row["state"] if sacct_row else None)
        if raw_state is None:
            return None

        mapped = SLURM_STATE_MAP.get(raw_state, "running")
        message = None
        if mapped == "failed":
            message = f"slurm state: {raw_state}"
        runtime = None
        if sacct_row:
            alloc_cpus = None
            if sacct_row["alloc_cpus"].isdigit():
                alloc_cpus = int(sacct_row["alloc_cpus"])
            runtime = {
                "slurm_state": raw_state,
                "slurm_elapsed": sacct_row["elapsed"] or None,
                "slurm_max_rss": sacct_row["max_rss"] or None,
                "slurm_alloc_cpus": alloc_cpus,
            }
        elif raw_state:
            runtime = {"slurm_state": raw_state}
        return BackendQueryResult(state=mapped, exit_code=None, message=message, runtime=runtime)

    def cancel(self, run_ctx: RunContext) -> None:
        return



from __future__ import annotations

import json
import os
import signal
from datetime import datetime
from pathlib import Path

from one_run.commands.common import load_run_paths
from one_run.errors import OneRunError
from one_run.models import RunMetadata, RunStatus
from one_run.payloads import cancel_result_payload
from one_run.run_dir import RunPaths
from one_run.utils.shell import command_exists, run_capture
from one_run.status_store import read_metadata, read_status, write_status
from one_run.tracking import append_event, write_summary

def _cancel_local(paths: RunPaths, status: RunStatus, *, force: bool = False) -> RunStatus:
    now = datetime.now()

    if paths.local_pid.exists():
        try:
            pid = int(paths.local_pid.read_text(encoding="utf-8").strip())
            os.kill(pid, signal.SIGKILL if force else signal.SIGTERM)
        except ProcessLookupError:
            pass
        except (ValueError, OSError) as exc:
            raise OneRunError(f"failed to cancel local run: {exc}") from exc

    exit_code = 137 if force else 130
    message = "force cancelled by user" if force else "cancelled by user"
    paths.local_exit_code.write_text(f"{exit_code}\n", encoding="utf-8")
    updated = status.model_copy(
        update={
            "state": "failed",
            "ended_at": now,
            "exit_code": exit_code,
            "message": message,
        }
    )
    write_status(paths.status, updated)
    append_event(paths, run_id=paths.run_dir.name, event="cancelled", state=updated.state, message=updated.message)
    write_summary(paths, paths.run_dir.name)
    return updated


def _cancel_slurm(metadata: RunMetadata, status: RunStatus, paths: RunPaths, *, force: bool = False) -> RunStatus:
    job_id = metadata.slurm_job_id
    if not job_id:
        raise OneRunError("slurm run has no job id")
    if not command_exists("scancel"):
        raise OneRunError("scancel not found. cannot cancel slurm job")

    cmd = ["scancel", "--signal=KILL", job_id] if force else ["scancel", job_id]
    proc = run_capture(cmd)
    if proc.returncode != 0:
        msg = (proc.stderr or proc.stdout).strip() or "scancel failed"
        raise OneRunError(f"failed to cancel slurm job {job_id}: {msg}")

    exit_code = 137 if force else 130
    message = "force cancelled by user" if force else "cancelled by user"
    updated = status.model_copy(
        update={
            "state": "failed",
            "ended_at": datetime.now(),
            "exit_code": exit_code,
            "message": message,
        }
    )
    write_status(paths.status, updated)
    append_event(paths, run_id=paths.run_dir.name, event="cancelled", state=updated.state, message=updated.message)
    write_summary(paths, paths.run_dir.name)
    return updated


def cancel_run(run_id: str, *, force: bool = False, json_output: bool = False) -> int:
    project_root = Path.cwd().resolve()
    paths = load_run_paths(project_root, run_id)
    metadata = RunMetadata.model_validate(read_metadata(paths.metadata))
    status = read_status(paths.status)

    if status.state in {"success", "failed"}:
        if json_output:
            print(json.dumps(cancel_result_payload(
                run_id=run_id,
                state=status.state,
                force=force,
                cancelled=False,
            )))
        else:
            print(f"run_id={run_id}")
            print(f"state={status.state}")
            print("cancelled=false")
        return 0

    if metadata.backend_type == "slurm":
        updated = _cancel_slurm(metadata, status, paths, force=force)
    else:
        updated = _cancel_local(paths, status, force=force)

    if json_output:
        print(json.dumps(cancel_result_payload(
            run_id=run_id,
            state=updated.state,
            force=force,
            cancelled=True,
            exit_code=updated.exit_code,
            message=updated.message,
        )))
    else:
        print(f"run_id={run_id}")
        print(f"state={updated.state}")
        print(f"force={str(force).lower()}")
        print("cancelled=true")
    return 0



from __future__ import annotations

from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field, model_validator
from one_run.constants import JSON_SCHEMA_VERSION


class ExperimentConfig(BaseModel):
    name: str = Field(min_length=1)
    seed: int
    tags: list[str] | None = None
    color: Literal["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"] | None = None


class EntrypointConfig(BaseModel):
    command: str = Field(min_length=1)


class EnvironmentConfig(BaseModel):
    kind: Literal["local", "docker"]
    image: str | None = None
    workdir: str | None = None
    mounts: list[str] | None = None
    env: dict[str, str] | None = None
    env_file: str | None = None
    pull_policy: Literal["always", "if_missing", "never"] | None = None
    labels: dict[str, str] | None = None

    @model_validator(mode="after")
    def validate_docker(self) -> "EnvironmentConfig":
        if self.kind == "docker" and not self.image:
            raise ValueError("environment.image is required when environment.kind=docker")
        return self


class ResourcesConfig(BaseModel):
    gpus: int | None = None
    cpus: int | None = None
    mem_gb: int | None = None
    time_min: int | None = None


class SlurmConfig(BaseModel):
    partition: str | None = None
    retry_max_attempts: int | None = None
    retry_backoff_sec: int | None = None
    retry_on_states: list[Literal["failed", "timeout", "cancelled"]] | None = None


class BackendConfig(BaseModel):
    kind: Literal["local", "slurm"]
    slurm: SlurmConfig | None = None


class ScheduleConfig(BaseModel):
    cron: str = Field(min_length=1)
    enabled: bool = True


class DependsOnConfig(BaseModel):
    run_ids: list[str] | None = None
    tags: list[str] | None = None
    timeout: str | None = None
    interval_sec: float = 2.0

    @model_validator(mode="after")
    def validate_depends_on(self) -> "DependsOnConfig":
        has_run_ids = bool(self.run_ids)
        has_tags = bool(self.tags)
        if not has_run_ids and not has_tags:
            raise ValueError("depends_on requires run_ids and/or tags")
        if self.interval_sec <= 0:
            raise ValueError("depends_on.interval_sec must be > 0")
        return self


class ListenerConfig(BaseModel):
    kind: Literal["local", "webhook"]
    name: str | None = None
    action: str | None = None
    events: list[str] | None = None
    states: list[str] | None = None
    tags: list[str] | None = None
    handler: str | None = None
    webhook_url: str | None = None
    timeout_sec: float | None = None

    @model_validator(mode="after")
    def validate_listener(self) -> "ListenerConfig":
        if self.kind == "webhook" and not self.webhook_url:
            raise ValueError("listeners[].webhook_url is required when kind=webhook")
        if self.timeout_sec is not None and self.timeout_sec <= 0:
            raise ValueError("listeners[].timeout_sec must be > 0")
        return self


class Manifest(BaseModel):
    experiment: ExperimentConfig
    entrypoint: EntrypointConfig
    environment: EnvironmentConfig
    resources: ResourcesConfig | None = None
    backend: BackendConfig
    schedule: ScheduleConfig | None = None
    depends_on: DependsOnConfig | None = None
    listeners: list[ListenerConfig] | None = None


class RunMetadata(BaseModel):
    schema_version: str = JSON_SCHEMA_VERSION
    run_id: str
    timestamp: datetime
    hostname: str
    cwd: str
    project_root: str
    git_commit: str | None
    git_dirty: bool | None
    experiment_name: str | None = None
    tags: list[str] | None = None
    experiment_color: Literal["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"] | None = None
    backend_type: Literal["local", "slurm"]
    docker_image: str | None
    docker_image_digest: str | None = None
    slurm_job_id: str | None = None
    slurm_attempt: int | None = None


class RunStatus(BaseModel):
    schema_version: str = JSON_SCHEMA_VERSION
    state: Literal["pending", "running", "success", "failed"]
    started_at: datetime | None = None
    ended_at: datetime | None = None
    exit_code: int | None = None
    message: str | None = None


class RunRuntime(BaseModel):
    schema_version: str = JSON_SCHEMA_VERSION
    collected_at: datetime
    duration_sec: float | None = None
    peak_rss_mb: float | None = None
    cpu_count: int | None = None
    total_ram_gb: float | None = None
    slurm_elapsed: str | None = None
    slurm_max_rss: str | None = None
    slurm_alloc_cpus: int | None = None
